@@ -31,10 +31,14 @@ bool pl_color_system_is_ycbcr_like(enum pl_color_system sys)
     case PL_COLOR_SYSTEM_SMPTE_240M:
     case PL_COLOR_SYSTEM_BT_2020_NC:
     case PL_COLOR_SYSTEM_BT_2020_C:
+    case PL_COLOR_SYSTEM_BT_2100_PQ:
+    case PL_COLOR_SYSTEM_BT_2100_HLG:
     case PL_COLOR_SYSTEM_YCGCO:
         return true;
-    default: abort();
+    case PL_COLOR_SYSTEM_COUNT: break;
     };
+
+    abort();
 }
 
 bool pl_color_system_is_linear(enum pl_color_system sys)
@@ -49,10 +53,14 @@ bool pl_color_system_is_linear(enum pl_color_system sys)
     case PL_COLOR_SYSTEM_YCGCO:
         return true;
     case PL_COLOR_SYSTEM_BT_2020_C:
+    case PL_COLOR_SYSTEM_BT_2100_PQ:
+    case PL_COLOR_SYSTEM_BT_2100_HLG:
     case PL_COLOR_SYSTEM_XYZ:
         return false;
-    default: abort();
+    case PL_COLOR_SYSTEM_COUNT: break;
     };
+
+    abort();
 }
 
 enum pl_color_system pl_color_system_guess_ycbcr(int width, int height)
@@ -181,11 +189,14 @@ bool pl_color_primaries_is_wide_gamut(enum pl_color_primaries prim)
     case PL_COLOR_PRIM_PRO_PHOTO:
     case PL_COLOR_PRIM_CIE_1931:
     case PL_COLOR_PRIM_DCI_P3:
+    case PL_COLOR_PRIM_DISPLAY_P3:
     case PL_COLOR_PRIM_V_GAMUT:
     case PL_COLOR_PRIM_S_GAMUT:
         return true;
-    default: abort();
+    case PL_COLOR_PRIM_COUNT: break;
     }
+
+    abort();
 }
 
 enum pl_color_primaries pl_color_primaries_guess(int width, int height)
@@ -224,8 +235,10 @@ float pl_color_transfer_nominal_peak(enum pl_color_transfer trc)
     case PL_COLOR_TRC_V_LOG:    return 46.0855;
     case PL_COLOR_TRC_S_LOG1:   return 6.52;
     case PL_COLOR_TRC_S_LOG2:   return 9.212;
-    default: abort();
+    case PL_COLOR_TRC_COUNT: break;
     }
+
+    abort();
 }
 
 bool pl_color_light_is_scene_referred(enum pl_color_light light)
@@ -238,8 +251,10 @@ bool pl_color_light_is_scene_referred(enum pl_color_light light)
     case PL_COLOR_LIGHT_SCENE_709_1886:
     case PL_COLOR_LIGHT_SCENE_1_2:
         return true;
-    default: abort();
+    case PL_COLOR_LIGHT_COUNT: break;
     }
+
+    abort();
 }
 
 const struct pl_color_space pl_color_space_unknown = {0};
@@ -530,6 +545,14 @@ struct pl_matrix3x3 pl_get_xyz2rgb_matrix(const struct pl_raw_primaries *prim)
     return out;
 }
 
+// LMS<-XYZ revised matrix from CIECAM97, based on a linear transform and
+// normalized for equal energy on monochrome inputs
+static const struct pl_matrix3x3 m_cat97 = {{
+    {  0.8562,  0.3372, -0.1934 },
+    { -0.8360,  1.8327,  0.0033 },
+    {  0.0357, -0.0469,  1.0112 },
+}};
+
 // M := M * XYZd<-XYZs
 static void apply_chromatic_adaptation(struct pl_cie_xy src,
                                        struct pl_cie_xy dest,
@@ -542,26 +565,19 @@ static void apply_chromatic_adaptation(struct pl_cie_xy src,
 
     // XYZd<-XYZs = Ma^-1 * (I*[Cd/Cs]) * Ma
     // http://www.brucelindbloom.com/index.html?Eqn_ChromAdapt.html
+    // For Ma, we use the CIECAM97 revised (linear) matrix
     float C[3][2];
 
-    // Ma = Bradford matrix, arguably most popular method in use today.
-    // This is derived experimentally and thus hard-coded.
-    struct pl_matrix3x3 bradford = {{
-        {  0.8951,  0.2664, -0.1614 },
-        { -0.7502,  1.7135,  0.0367 },
-        {  0.0389, -0.0685,  1.0296 },
-    }};
-
     for (int i = 0; i < 3; i++) {
         // source cone
-        C[i][0] = bradford.m[i][0] * pl_cie_X(src)
-                + bradford.m[i][1] * 1
-                + bradford.m[i][2] * pl_cie_Z(src);
+        C[i][0] = m_cat97.m[i][0] * pl_cie_X(src)
+                + m_cat97.m[i][1] * 1
+                + m_cat97.m[i][2] * pl_cie_Z(src);
 
         // dest cone
-        C[i][1] = bradford.m[i][0] * pl_cie_X(dest)
-                + bradford.m[i][1] * 1
-                + bradford.m[i][2] * pl_cie_Z(dest);
+        C[i][1] = m_cat97.m[i][0] * pl_cie_X(dest)
+                + m_cat97.m[i][1] * 1
+                + m_cat97.m[i][2] * pl_cie_Z(dest);
     }
 
     // tmp := I * [Cd/Cs] * Ma
@@ -569,14 +585,159 @@ static void apply_chromatic_adaptation(struct pl_cie_xy src,
     for (int i = 0; i < 3; i++)
         tmp.m[i][i] = C[i][1] / C[i][0];
 
-    pl_matrix3x3_mul(&tmp, &bradford);
+    pl_matrix3x3_mul(&tmp, &m_cat97);
 
     // M := M * Ma^-1 * tmp
-    pl_matrix3x3_invert(&bradford);
-    pl_matrix3x3_mul(mat, &bradford);
+    struct pl_matrix3x3 ma_inv = m_cat97;
+    pl_matrix3x3_invert(&ma_inv);
+    pl_matrix3x3_mul(mat, &ma_inv);
     pl_matrix3x3_mul(mat, &tmp);
 }
 
+const struct pl_cone_params pl_vision_normal        = {PL_CONE_NONE, 1.0};
+const struct pl_cone_params pl_vision_protanomaly   = {PL_CONE_L,    0.5};
+const struct pl_cone_params pl_vision_protanopia    = {PL_CONE_L,    0.0};
+const struct pl_cone_params pl_vision_deuteranomaly = {PL_CONE_M,    0.5};
+const struct pl_cone_params pl_vision_deuteranopia  = {PL_CONE_M,    0.0};
+const struct pl_cone_params pl_vision_tritanomaly   = {PL_CONE_S,    0.5};
+const struct pl_cone_params pl_vision_tritanopia    = {PL_CONE_S,    0.0};
+const struct pl_cone_params pl_vision_monochromacy  = {PL_CONE_LM,   0.0};
+const struct pl_cone_params pl_vision_achromatopsia = {PL_CONE_LMS,  0.0};
+
+struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
+                                       const struct pl_raw_primaries *prim)
+{
+    // LMS<-RGB := LMS<-XYZ * XYZ<-RGB
+    struct pl_matrix3x3 rgb2lms = m_cat97;
+    struct pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(prim);
+    pl_matrix3x3_mul(&rgb2lms, &rgb2xyz);
+
+    // LMS versions of the two opposing primaries, plus neutral
+    float lms_r[3] = {1.0, 0.0, 0.0},
+          lms_b[3] = {0.0, 0.0, 1.0},
+          lms_w[3] = {1.0, 1.0, 1.0};
+
+    pl_matrix3x3_apply(&rgb2lms, lms_r);
+    pl_matrix3x3_apply(&rgb2lms, lms_b);
+    pl_matrix3x3_apply(&rgb2lms, lms_w);
+
+    float a, b, c = params->strength;
+    struct pl_matrix3x3 distort;
+
+    switch (params->cones) {
+    case PL_CONE_NONE:
+        return pl_matrix3x3_identity;
+
+    case PL_CONE_L:
+        // Solve to preserve neutral and blue
+        a = (lms_b[0] - lms_b[2] * lms_w[0] / lms_w[2]) /
+            (lms_b[1] - lms_b[2] * lms_w[1] / lms_w[2]);
+        b = (lms_b[0] - lms_b[1] * lms_w[0] / lms_w[1]) /
+            (lms_b[2] - lms_b[1] * lms_w[2] / lms_w[1]);
+        assert(fabs(a * lms_w[1] + b * lms_w[2] - lms_w[0]) < 1e-6);
+
+        distort = (struct pl_matrix3x3) {{
+            {            c, (1.0 - c) * a, (1.0 - c) * b},
+            {          0.0,           1.0,           0.0},
+            {          0.0,           0.0,           1.0},
+        }};
+        break;
+
+    case PL_CONE_M:
+        // Solve to preserve neutral and blue
+        a = (lms_b[1] - lms_b[2] * lms_w[1] / lms_w[2]) /
+            (lms_b[0] - lms_b[2] * lms_w[0] / lms_w[2]);
+        b = (lms_b[1] - lms_b[0] * lms_w[1] / lms_w[0]) /
+            (lms_b[2] - lms_b[0] * lms_w[2] / lms_w[0]);
+        assert(fabs(a * lms_w[0] + b * lms_w[2] - lms_w[1]) < 1e-6);
+
+        distort = (struct pl_matrix3x3) {{
+            {          1.0,           0.0,           0.0},
+            {(1.0 - c) * a,             c, (1.0 - c) * b},
+            {          0.0,           0.0,           1.0},
+        }};
+        break;
+
+    case PL_CONE_S:
+        // Solve to preserve neutral and red
+        a = (lms_r[2] - lms_r[1] * lms_w[2] / lms_w[1]) /
+            (lms_r[0] - lms_r[1] * lms_w[0] / lms_w[1]);
+        b = (lms_r[2] - lms_r[0] * lms_w[2] / lms_w[0]) /
+            (lms_r[1] - lms_r[0] * lms_w[1] / lms_w[0]);
+        assert(fabs(a * lms_w[0] + b * lms_w[1] - lms_w[2]) < 1e-6);
+
+        distort = (struct pl_matrix3x3) {{
+            {          1.0,           0.0,           0.0},
+            {          0.0,           1.0,           0.0},
+            {(1.0 - c) * a, (1.0 - c) * b,             c},
+        }};
+        break;
+
+    case PL_CONE_LM:
+        // Solve to preserve neutral
+        a = lms_w[0] / lms_w[2];
+        b = lms_w[1] / lms_w[2];
+
+        distort = (struct pl_matrix3x3) {{
+            {            c,           0.0, (1.0 - c) * a},
+            {          0.0,             c, (1.0 - c) * b},
+            {          0.0,           0.0,           1.0},
+        }};
+        break;
+
+    case PL_CONE_MS:
+        // Solve to preserve neutral
+        a = lms_w[1] / lms_w[0];
+        b = lms_w[2] / lms_w[0];
+
+        distort = (struct pl_matrix3x3) {{
+            {          1.0,           0.0,           0.0},
+            {(1.0 - c) * a,             c,           0.0},
+            {(1.0 - c) * b,           0.0,             c},
+        }};
+        break;
+
+    case PL_CONE_LS:
+        // Solve to preserve neutral
+        a = lms_w[0] / lms_w[1];
+        b = lms_w[2] / lms_w[1];
+
+        distort = (struct pl_matrix3x3) {{
+            {            c, (1.0 - c) * a,           0.0},
+            {          0.0,           1.0,           0.0},
+            {          0.0, (1.0 - c) * b,             c},
+        }};
+        break;
+
+    case PL_CONE_LMS: {
+        // Rod cells only, which can be modelled somewhat as a combination of
+        // L and M cones. Either way, this is pushing the limits of the our
+        // color model, so this is only a rough approximation.
+        const float w[3] = {0.3605, 0.6415, -0.002};
+        assert(fabs(w[0] + w[1] + w[2] - 1.0) < 1e-6);
+
+        for (int i = 0; i < 3; i++) {
+            for (int j = 0; j < 3; j++) {
+                distort.m[i][j] = (1.0 - c) * w[j] * lms_w[i] / lms_w[j];
+                if (i == j)
+                    distort.m[i][j] += c;
+            }
+        }
+        break;
+    }
+
+    default: abort();
+    }
+
+    // out := RGB<-LMS * distort * LMS<-RGB
+    struct pl_matrix3x3 out = rgb2lms;
+    pl_matrix3x3_invert(&out);
+    pl_matrix3x3_mul(&out, &distort);
+    pl_matrix3x3_mul(&out, &rgb2lms);
+
+    return out;
+}
+
 struct pl_matrix3x3 pl_get_color_mapping_matrix(const struct pl_raw_primaries *src,
                                                 const struct pl_raw_primaries *dst,
                                                 enum pl_rendering_intent intent)
@@ -653,6 +814,16 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
             {0, 1, 0}
         }};
         break;
+    case PL_COLOR_SYSTEM_BT_2100_PQ:
+    case PL_COLOR_SYSTEM_BT_2100_HLG:
+        // This is the ICtCp<-LMS matrix from the spec, so invert it
+        m = (struct pl_matrix3x3) {{
+            { 2048/4096.,   2048/4096.,          0},
+            { 6610/4096., -13613/4096., 7003/4096.},
+            {17933/4096., -17390/4096., -543/4096.},
+        }};
+        pl_matrix3x3_invert(&m);
+        break;
     case PL_COLOR_SYSTEM_YCGCO:
         m = (struct pl_matrix3x3) {{
             {1,  -1,  1},
@@ -668,8 +839,8 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         // For lack of anything saner to do, just assume the caller wants
         // BT.709 primaries, which is a reasonable assumption.
         m = pl_get_xyz2rgb_matrix(pl_raw_primaries_get(PL_COLOR_PRIM_BT_709));
-    }
         break;
+    }
     default: abort();
     }
 
