@@ -24,10 +24,13 @@
 #include <stdint.h>
 #include <inttypes.h>
 
-#include "ta/talloc.h"
 #include "config.h"
+#include "config_internal.h"
 #include "pl_assert.h"
 
+#include <xtalloc.h>
+#include <bstr.h>
+
 // Include all of the symbols that should be public in a way that marks them
 // as being externally visible. (Otherwise, all symbols are hidden by default)
 #pragma GCC visibility push(default)
@@ -41,6 +44,7 @@
 #include "include/libplacebo/gpu.h"
 #include "include/libplacebo/renderer.h"
 #include "include/libplacebo/shaders.h"
+#include "include/libplacebo/shaders/av1.h"
 #include "include/libplacebo/shaders/colorspace.h"
 #include "include/libplacebo/shaders/sampling.h"
 #include "include/libplacebo/swapchain.h"
@@ -62,6 +66,9 @@
 // Returns the log base 2 of an unsigned long long
 #define PL_LOG2(x) ((unsigned) (8*sizeof (unsigned long long) - __builtin_clzll((x)) - 1))
 
+// Returns whether or not a number is a power of two (or zero)
+#define PL_ISPOT(x) (((x) & ((x) - 1)) == 0)
+
 // Returns the size of a static array with known size.
 #define PL_ARRAY_SIZE(s) (sizeof(s) / sizeof((s)[0]))
 
@@ -102,6 +109,7 @@ static inline float *pl_transpose(int dim, float *out, const float *in)
 // Helpers for doing alignment calculations
 static inline size_t pl_gcd(size_t x, size_t y)
 {
+    assert(x && y);
     while (y) {
         size_t tmp = y;
         y = x % y;
@@ -113,5 +121,6 @@ static inline size_t pl_gcd(size_t x, size_t y)
 
 static inline size_t pl_lcm(size_t x, size_t y)
 {
+    assert(x && y);
     return x * (y / pl_gcd(x, y));
 }
