@@ -18,6 +18,7 @@
 #include "common.h"
 #include "context.h"
 #include "shaders.h"
+#include "dispatch.h"
 #include "gpu.h"
 
 enum {
@@ -70,7 +71,7 @@ struct pass {
 
     // for uniform buffer updates
     const struct pl_buf *ubo;
-    struct pl_desc ubo_desc; // temporary
+    struct pl_shader_desc ubo_desc; // temporary
 
     // Cached pl_pass_run_params. This will also contain mutable allocations
     // for the push constants, descriptor bindings (including the binding for
@@ -114,9 +115,9 @@ void pl_dispatch_destroy(struct pl_dispatch **ptr)
     *ptr = NULL;
 }
 
-struct pl_shader *pl_dispatch_begin(struct pl_dispatch *dp)
+struct pl_shader *pl_dispatch_begin_ex(struct pl_dispatch *dp, bool unique)
 {
-    uint8_t ident = dp->current_ident++;
+    uint8_t ident = unique ? dp->current_ident++ : 0;
 
     struct pl_shader *sh;
     if (TARRAY_POP(dp->shaders, dp->num_shaders, &sh)) {
@@ -124,7 +125,7 @@ struct pl_shader *pl_dispatch_begin(struct pl_dispatch *dp)
         return sh;
     }
 
-    return pl_shader_alloc(dp->ctx, dp->gpu, ident, dp->current_index);
+    return pl_shader_alloc_ex(dp->ctx, dp->gpu, dp->current_index, ident);
 }
 
 void pl_dispatch_reset_frame(struct pl_dispatch *dp)
@@ -133,6 +134,11 @@ void pl_dispatch_reset_frame(struct pl_dispatch *dp)
     dp->current_index++;
 }
 
+struct pl_shader *pl_dispatch_begin(struct pl_dispatch *dp)
+{
+    return pl_dispatch_begin_ex(dp, false);
+}
+
 static bool add_pass_var(struct pl_dispatch *dp, void *tmp, struct pass *pass,
                          struct pl_pass_params *params,
                          const struct pl_shader_var *sv, struct pass_var *pv)
@@ -143,7 +149,7 @@ static bool add_pass_var(struct pl_dispatch *dp, void *tmp, struct pass *pass,
     // this is likely to exceed the VGPR/pushc size budgets
     bool try_pushc = (sv->var.dim_m == 1 && sv->var.dim_a == 1) || sv->dynamic;
     if (try_pushc && gpu->glsl.vulkan && gpu->limits.max_pushc_size) {
-        pv->layout = pl_push_constant_layout(gpu, params->push_constants_size, &sv->var);
+        pv->layout = pl_std430_layout(params->push_constants_size, &sv->var);
         size_t new_size = pv->layout.offset + pv->layout.size;
         if (new_size <= gpu->limits.max_pushc_size) {
             params->push_constants_size = new_size;
@@ -160,7 +166,7 @@ static bool add_pass_var(struct pl_dispatch *dp, void *tmp, struct pass *pass,
     // the UBO writes every frame
     bool try_ubo = !(gpu->caps & PL_GPU_CAP_INPUT_VARIABLES) || !sv->dynamic;
     if (try_ubo && gpu->glsl.version >= 440 && gpu->limits.max_ubo_size) {
-        if (pl_buf_desc_append(tmp, gpu, &pass->ubo_desc, &pv->layout, sv->var)) {
+        if (sh_buf_desc_append(tmp, gpu, &pass->ubo_desc, &pv->layout, sv->var)) {
             pv->type = PASS_VAR_UBO;
             return true;
         }
@@ -267,17 +273,17 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
             const char *name = sva->attr.name;
 
             char loc[32];
-            snprintf(loc, sizeof(loc), "layout(location=%d) ", va->location);
-            ADD(vert_head, "%s%s %s %s;\n", loc, vert_in, type, va->name);
+            snprintf(loc, sizeof(loc), "layout(location=%d)", va->location);
+            ADD(vert_head, "%s %s %s %s;\n", loc, vert_in, type, va->name);
 
             if (strcmp(name, vert_pos) == 0) {
                 pl_assert(va->fmt->num_components == 2);
                 ADD(vert_body, "gl_Position = vec4(%s, 0.0, 1.0);\n", va->name);
             } else {
                 // Everything else is just blindly passed through
-                ADD(vert_head, "%s%s %s %s;\n", loc, vert_out, type, name);
+                ADD(vert_head, "%s %s %s %s;\n", loc, vert_out, type, name);
                 ADD(vert_body, "%s = %s;\n", name, va->name);
-                ADD(glsl, "%s%s %s %s;\n", loc, frag_in, type, name);
+                ADD(glsl, "%s %s %s %s;\n", loc, frag_in, type, name);
             }
         }
 
@@ -307,7 +313,7 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
             struct pass_var *pv = &pass->vars[i];
             if (pv->type != PASS_VAR_PUSHC)
                 continue;
-            ADD(glsl, "/*offset=%zu*/ ", pv->layout.offset);
+            ADD(glsl, "    layout(offset=%zu) ", pv->layout.offset);
             add_var(dp, glsl, var);
         }
         ADD(glsl, "};\n");
@@ -364,13 +370,13 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
         case PL_DESC_BUF_UNIFORM:
             ADD(glsl, "layout(std140, binding=%d) uniform %s ", desc->binding,
                 desc->name);
-            add_buffer_vars(dp, glsl, desc->buffer_vars, desc->num_buffer_vars);
+            add_buffer_vars(dp, glsl, sd->buffer_vars, sd->num_buffer_vars);
             break;
 
         case PL_DESC_BUF_STORAGE:
             ADD(glsl, "layout(std430, binding=%d) %s buffer %s ", desc->binding,
                 pl_desc_access_glsl_name(desc->access), desc->name);
-            add_buffer_vars(dp, glsl, desc->buffer_vars, desc->num_buffer_vars);
+            add_buffer_vars(dp, glsl, sd->buffer_vars, sd->num_buffer_vars);
             break;
 
         case PL_DESC_BUF_TEXEL_UNIFORM:
@@ -449,8 +455,21 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
 
     for (int i = 0; i < dp->num_passes; i++) {
         const struct pass *p = dp->passes[i];
-        if (p->signature == sig && blend_equal(p->pass->params.blend_params, blend))
+        if (p->signature != sig)
+            continue;
+
+        if (pl_shader_is_compute(sh)) {
+            // no special requirements besides the signature
             return dp->passes[i];
+        } else {
+            pl_assert(target);
+            const struct pl_fmt *tfmt;
+            tfmt = p->pass->params.target_dummy.params.format;
+            bool raster_ok = target->params.format == tfmt;
+            raster_ok &= blend_equal(p->pass->params.blend_params, blend);
+            if (raster_ok)
+                return dp->passes[i];
+        }
     }
 
     void *tmp = talloc_new(NULL); // for resources attached to `params`
@@ -458,9 +477,11 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
     struct pass *pass = talloc_zero(dp, struct pass);
     pass->signature = sig;
     pass->failed = true; // will be set to false on success
-    pass->ubo_desc = (struct pl_desc) {
-        .name = "UBO",
-        .type = PL_DESC_BUF_UNIFORM,
+    pass->ubo_desc = (struct pl_shader_desc) {
+        .desc = {
+            .name = "UBO",
+            .type = PL_DESC_BUF_UNIFORM,
+        },
     };
 
     struct pl_shader_res *res = &sh->res;
@@ -473,6 +494,7 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
     };
 
     if (params.type == PL_PASS_RASTER) {
+        assert(target);
         params.target_dummy = *target;
 
         // Fill in the vertex attributes array
@@ -517,7 +539,7 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
 
     // Create and attach the UBO if necessary
     int ubo_index = -1;
-    size_t ubo_size = pl_buf_desc_size(&pass->ubo_desc);
+    size_t ubo_size = sh_buf_desc_size(&pass->ubo_desc);
     if (ubo_size) {
         pass->ubo = pl_buf_create(dp->gpu, &(struct pl_buf_params) {
             .type = PL_BUF_UNIFORM,
@@ -531,10 +553,8 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
         }
 
         ubo_index = res->num_descriptors;
-        sh_desc(sh, (struct pl_shader_desc) {
-            .desc = pass->ubo_desc,
-            .object = pass->ubo,
-        });
+        pass->ubo_desc.object = pass->ubo;
+        sh_desc(sh, pass->ubo_desc);
     }
 
     // Place and fill in the descriptors
@@ -570,7 +590,7 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
     pass->failed = false;
 
 error:
-    pass->ubo_desc = (struct pl_desc) {0}; // contains temporary pointers
+    pass->ubo_desc = (struct pl_shader_desc) {0}; // contains temporary pointers
     talloc_free(tmp);
     TARRAY_APPEND(dp, dp->passes, dp->num_passes, pass);
     return pass;
@@ -715,6 +735,11 @@ bool pl_dispatch_finish(struct pl_dispatch *dp, struct pl_shader **psh,
     const struct pl_shader_res *res = &sh->res;
     bool ret = false;
 
+    if (sh->failed) {
+        PL_ERR(sh, "Trying to dispatch a failed shader.");
+        goto error;
+    }
+
     if (!sh->mutable) {
         PL_ERR(dp, "Trying to dispatch non-mutable shader?");
         goto error;
@@ -727,11 +752,17 @@ bool pl_dispatch_finish(struct pl_dispatch *dp, struct pl_shader **psh,
 
     const struct pl_tex_params *tpars = &target->params;
     if (pl_tex_params_dimension(*tpars) != 2 || !tpars->renderable) {
-        PL_ERR(dp, "Trying to dispatch using a shader using an invalid target "
+        PL_ERR(dp, "Trying to dispatch a shader using an invalid target "
                "texture. The target must be a renderable 2D texture.");
         goto error;
     }
 
+    if (pl_shader_is_compute(sh) && !tpars->storable) {
+        PL_ERR(dp, "Trying to dispatch using a compute shader with a "
+               "non-storable target texture.");
+        goto error;
+    }
+
     struct pl_rect2d full = {0, 0, tpars->w, tpars->h};
     rc = PL_DEF(rc, &full);
 
@@ -831,6 +862,11 @@ bool pl_dispatch_compute(struct pl_dispatch *dp, struct pl_shader **psh,
     const struct pl_shader_res *res = &sh->res;
     bool ret = false;
 
+    if (sh->failed) {
+        PL_ERR(sh, "Trying to dispatch a failed shader.");
+        goto error;
+    }
+
     if (!sh->mutable) {
         PL_ERR(dp, "Trying to dispatch non-mutable shader?");
         goto error;
