@@ -69,6 +69,26 @@ void pl_gpu_print_info(const struct pl_gpu *gpu, enum pl_log_level lev)
     LOG(PRIu32, align_tex_xfer_stride);
     LOG("zu", align_tex_xfer_offset);
 #undef LOG
+
+    if (pl_gpu_supports_interop(gpu)) {
+        PL_MSG(gpu, lev, "    External API interop:");
+
+        // Pretty-print the device UUID
+        static const char *hexdigits = "0123456789ABCDEF";
+        char buf[3 * sizeof(gpu->uuid)];
+        for (int i = 0; i < sizeof(gpu->uuid); i++) {
+            uint8_t x = gpu->uuid[i];
+            buf[3 * i + 0] = hexdigits[x >> 4];
+            buf[3 * i + 1] = hexdigits[x & 0xF];
+            buf[3 * i + 2] = i == sizeof(gpu->uuid) - 1 ? '\0' : ':';
+        }
+
+        PL_MSG(gpu, lev, "      UUID: %s", buf);
+        PL_MSG(gpu, lev, "      shared memory caps: 0x%x",
+               (unsigned int) gpu->handle_caps.shared_mem);
+        PL_MSG(gpu, lev, "      sync caps: 0x%x",
+               (unsigned int) gpu->handle_caps.sync);
+    }
 }
 
 static int cmp_fmt(const void *pa, const void *pb)
@@ -334,6 +354,11 @@ const struct pl_fmt *pl_find_named_fmt(const struct pl_gpu *gpu, const char *nam
 const struct pl_tex *pl_tex_create(const struct pl_gpu *gpu,
                                    const struct pl_tex_params *params)
 {
+    if (params->handle_type) {
+        pl_assert(params->handle_type & gpu->handle_caps.shared_mem);
+        pl_assert(PL_ISPOT(params->handle_type));
+    }
+
     switch (pl_tex_params_dimension(*params)) {
     case 1:
         pl_assert(params->w > 0);
@@ -384,8 +409,15 @@ static bool pl_tex_params_superset(struct pl_tex_params a, struct pl_tex_params
 bool pl_tex_recreate(const struct pl_gpu *gpu, const struct pl_tex **tex,
                      const struct pl_tex_params *params)
 {
-    if (*tex && pl_tex_params_superset((*tex)->params, *params))
+    if (params->initial_data) {
+        PL_ERR(gpu, "pl_tex_recreate may not be used with `initial_data`!");
+        return false;
+    }
+
+    if (*tex && pl_tex_params_superset((*tex)->params, *params)) {
+        pl_tex_invalidate(gpu, *tex);
         return true;
+    }
 
     PL_INFO(gpu, "(Re)creating %dx%dx%d texture", params->w, params->h, params->d);
     pl_tex_destroy(gpu, tex);
@@ -414,7 +446,8 @@ void pl_tex_clear(const struct pl_gpu *gpu, const struct pl_tex *dst,
 
 void pl_tex_invalidate(const struct pl_gpu *gpu, const struct pl_tex *tex)
 {
-    gpu->impl->tex_invalidate(gpu, tex);
+    if (gpu->impl->tex_invalidate)
+        gpu->impl->tex_invalidate(gpu, tex);
 }
 
 static void strip_coords(const struct pl_tex *tex, struct pl_rect3d *rc)
@@ -547,7 +580,6 @@ static void fix_tex_transfer(const struct pl_gpu *gpu,
         const struct pl_buf *buf = params->buf;
         size_t size = pl_tex_transfer_size(params);
         size_t texel = tex->params.format->texel_size;
-        pl_assert(params->buf_offset == PL_ALIGN2(params->buf_offset, 4));
         pl_assert(params->buf_offset == PL_ALIGN(params->buf_offset, texel));
         pl_assert(params->buf_offset + size <= buf->params.size);
     }
@@ -581,6 +613,11 @@ bool pl_tex_download(const struct pl_gpu *gpu,
 const struct pl_buf *pl_buf_create(const struct pl_gpu *gpu,
                                    const struct pl_buf_params *params)
 {
+    if (params->handle_type) {
+        pl_assert(params->handle_type & gpu->handle_caps.shared_mem);
+        pl_assert(PL_ISPOT(params->handle_type));
+    }
+
     switch (params->type) {
     case PL_BUF_TEX_TRANSFER:
         pl_assert(gpu->limits.max_xfer_size);
@@ -619,6 +656,34 @@ const struct pl_buf *pl_buf_create(const struct pl_gpu *gpu,
     return buf;
 }
 
+static bool pl_buf_params_superset(struct pl_buf_params a, struct pl_buf_params b)
+{
+    return a.type            == b.type &&
+           a.format          == b.format &&
+           a.size            >= b.size &&
+           (a.host_mapped    || !b.host_mapped) &&
+           (a.host_writable  || !b.host_writable) &&
+           (a.host_readable  || !b.host_readable);
+}
+
+bool pl_buf_recreate(const struct pl_gpu *gpu, const struct pl_buf **buf,
+                     const struct pl_buf_params *params)
+{
+    if (params->initial_data) {
+        PL_ERR(gpu, "pl_buf_recreate may not be used with `initial_data`!");
+        return false;
+    }
+
+    if (*buf && pl_buf_params_superset((*buf)->params, *params))
+        return true;
+
+    PL_INFO(gpu, "(Re)creating %zu buffer", params->size);
+    pl_buf_destroy(gpu, buf);
+    *buf = pl_buf_create(gpu, params);
+
+    return !!*buf;
+}
+
 void pl_buf_destroy(const struct pl_gpu *gpu, const struct pl_buf **buf)
 {
     if (!*buf)
@@ -646,6 +711,12 @@ bool pl_buf_read(const struct pl_gpu *gpu, const struct pl_buf *buf,
     return gpu->impl->buf_read(gpu, buf, buf_offset, dest, size);
 }
 
+bool pl_buf_export(const struct pl_gpu *gpu, const struct pl_buf *buf)
+{
+    pl_assert(buf->params.handle_type);
+    return gpu->impl->buf_export(gpu, buf);
+}
+
 bool pl_buf_poll(const struct pl_gpu *gpu, const struct pl_buf *buf, uint64_t t)
 {
     return gpu->impl->buf_poll ? gpu->impl->buf_poll(gpu, buf, t) : false;
@@ -751,71 +822,48 @@ struct pl_var_layout pl_var_host_layout(size_t offset, const struct pl_var *var)
     };
 }
 
-struct pl_var_layout pl_buf_uniform_layout(const struct pl_gpu *gpu, size_t offset,
-                                           const struct pl_var *var)
+struct pl_var_layout pl_std140_layout(size_t offset, const struct pl_var *var)
 {
-    if (gpu->limits.max_ubo_size) {
-        return gpu->impl->buf_uniform_layout(gpu, offset, var);
-    } else {
-        return (struct pl_var_layout) {0};
-    }
+    size_t el_size = pl_var_type_size(var->type);
+
+    // std140 packing rules:
+    // 1. The size of generic values is their size in bytes
+    // 2. The size of vectors is the vector length * the base count
+    // 3. Matrices are treated like arrays of column vectors
+    // 4. The size of array rows is that of the element size rounded up to
+    // the nearest multiple of vec4
+    // 5. All values are aligned to a multiple of their size (stride for arrays),
+    // with the exception of vec3 which is aligned like vec4
+    size_t stride = el_size * var->dim_v;
+    size_t align = stride;
+    if (var->dim_v == 3)
+        align += el_size;
+    if (var->dim_m * var->dim_a > 1)
+        stride = align = PL_ALIGN2(stride, sizeof(float[4]));
+
+    return (struct pl_var_layout) {
+        .offset = PL_ALIGN2(offset, align),
+        .stride = stride,
+        .size   = stride * var->dim_m * var->dim_a,
+    };
 }
 
-struct pl_var_layout pl_buf_storage_layout(const struct pl_gpu *gpu, size_t offset,
-                                           const struct pl_var *var)
+struct pl_var_layout pl_std430_layout(size_t offset, const struct pl_var *var)
 {
-    if (gpu->limits.max_ssbo_size) {
-        return gpu->impl->buf_storage_layout(gpu, offset, var);
-    } else {
-        return (struct pl_var_layout) {0};
-    }
-}
+    size_t el_size = pl_var_type_size(var->type);
 
-struct pl_var_layout pl_push_constant_layout(const struct pl_gpu *gpu, size_t offset,
-                                             const struct pl_var *var)
-{
-    if (gpu->limits.max_pushc_size) {
-        return gpu->impl->push_constant_layout(gpu, offset, var);
-    } else {
-        return (struct pl_var_layout) {0};
-    }
-}
+    // std430 packing rules: like std140, except arrays/matrices are always
+    // "tightly" packed, even arrays/matrices of vec3s
+    size_t stride = el_size * var->dim_v;
+    size_t align = stride;
+    if (var->dim_v == 3 && var->dim_m == 1 && var->dim_a == 1)
+        align += el_size;
 
-bool pl_buf_desc_append(void *tactx, const struct pl_gpu *gpu,
-                        struct pl_desc *buf_desc,
-                        struct pl_var_layout *out_layout,
-                        const struct pl_var new_var)
-{
-    struct pl_buffer_var bv = { .var = new_var };
-    size_t cur_size = pl_buf_desc_size(buf_desc);
-
-    switch (buf_desc->type) {
-    case PL_DESC_BUF_UNIFORM:
-        bv.layout = pl_buf_uniform_layout(gpu, cur_size, &new_var);
-        if (bv.layout.offset + bv.layout.size > gpu->limits.max_ubo_size)
-            return false;
-        break;
-    case PL_DESC_BUF_STORAGE:
-        bv.layout = pl_buf_storage_layout(gpu, cur_size, &new_var);
-        if (bv.layout.offset + bv.layout.size > gpu->limits.max_ssbo_size)
-            return false;
-        break;
-    default: abort();
-    }
-
-    *out_layout = bv.layout;
-    TARRAY_APPEND(tactx, buf_desc->buffer_vars, buf_desc->num_buffer_vars, bv);
-    return true;
-}
-
-size_t pl_buf_desc_size(const struct pl_desc *buf_desc)
-{
-    if (!buf_desc->num_buffer_vars)
-        return 0;
-
-    const struct pl_buffer_var *last;
-    last = &buf_desc->buffer_vars[buf_desc->num_buffer_vars - 1];
-    return last->layout.offset + last->layout.size;
+    return (struct pl_var_layout) {
+        .offset = PL_ALIGN2(offset, align),
+        .stride = stride,
+        .size   = stride * var->dim_m * var->dim_a,
+    };
 }
 
 void memcpy_layout(void *dst_p, struct pl_var_layout dst_layout,
@@ -909,21 +957,6 @@ void pl_pass_destroy(const struct pl_gpu *gpu, const struct pl_pass **pass)
     *pass = NULL;
 }
 
-static bool pl_tex_params_compat(const struct pl_tex_params a,
-                                 const struct pl_tex_params b)
-{
-    return a.format         == b.format &&
-           a.sampleable     == b.sampleable &&
-           a.renderable     == b.renderable &&
-           a.storable       == b.storable &&
-           a.blit_src       == b.blit_src &&
-           a.blit_dst       == b.blit_dst &&
-           a.host_writable  == b.host_writable &&
-           a.host_readable  == b.host_readable &&
-           a.sample_mode    == b.sample_mode &&
-           a.address_mode   == b.address_mode;
-}
-
 void pl_pass_run(const struct pl_gpu *gpu, const struct pl_pass_run_params *params)
 {
     const struct pl_pass *pass = params->pass;
@@ -993,7 +1026,7 @@ void pl_pass_run(const struct pl_gpu *gpu, const struct pl_pass_run_params *para
         const struct pl_tex *tex = params->target;
         pl_assert(tex);
         pl_assert(pl_tex_params_dimension(tex->params) == 2);
-        pl_assert(pl_tex_params_compat(tex->params, pass->params.target_dummy.params));
+        pl_assert(tex->params.format == pass->params.target_dummy.params.format);
         pl_assert(tex->params.renderable);
         struct pl_rect2d *vp = &new.viewport;
         struct pl_rect2d *sc = &new.scissors;
@@ -1048,52 +1081,13 @@ void pl_gpu_flush(const struct pl_gpu *gpu)
         gpu->impl->gpu_flush(gpu);
 }
 
+void pl_gpu_finish(const struct pl_gpu *gpu)
+{
+    gpu->impl->gpu_finish(gpu);
+}
+
 // GPU-internal helpers
 
-struct pl_var_layout std140_layout(const struct pl_gpu *gpu, size_t offset,
-                                   const struct pl_var *var)
-{
-    size_t el_size = pl_var_type_size(var->type);
-
-    // std140 packing rules:
-    // 1. The size of generic values is their size in bytes
-    // 2. The size of vectors is the vector length * the base count, with the
-    // exception of *vec3 which is always the same size as *vec4
-    // 3. Matrices are treated like arrays of column vectors
-    // 4. The size of array rows is that of the element size rounded up to
-    // the nearest multiple of vec4
-    // 5. All values are aligned to a multiple of their size (stride for arrays)
-    size_t size = el_size * var->dim_v;
-    if (var->dim_v == 3)
-        size += el_size;
-    if (var->dim_m * var->dim_a > 1)
-        size = PL_ALIGN2(size, sizeof(float[4]));
-
-    return (struct pl_var_layout) {
-        .offset = PL_ALIGN2(offset, size),
-        .stride = size,
-        .size   = size * var->dim_m * var->dim_a,
-    };
-}
-
-struct pl_var_layout std430_layout(const struct pl_gpu *gpu, size_t offset,
-                                   const struct pl_var *var)
-{
-    size_t el_size = pl_var_type_size(var->type);
-
-    // std430 packing rules: like std140, except arrays/matrices are always
-    // "tightly" packed, even arrays/matrices of vec3s
-    size_t size = el_size * var->dim_v;
-    if (var->dim_v == 3 && var->dim_m == 1 && var->dim_a == 1)
-        size += el_size;
-
-    return (struct pl_var_layout) {
-        .offset = PL_ALIGN2(offset, size),
-        .stride = size,
-        .size   = size * var->dim_m * var->dim_a,
-    };
-}
-
 void pl_buf_pool_uninit(const struct pl_gpu *gpu, struct pl_buf_pool *pool)
 {
     for (int i = 0; i < pool->num_buffers; i++)
@@ -1103,16 +1097,6 @@ void pl_buf_pool_uninit(const struct pl_gpu *gpu, struct pl_buf_pool *pool)
     *pool = (struct pl_buf_pool) {0};
 }
 
-static bool pl_buf_params_compatible(const struct pl_buf_params *new,
-                                     const struct pl_buf_params *old)
-{
-    return new->type == old->type &&
-           new->size <= old->size &&
-           new->host_mapped  == old->host_mapped &&
-           new->host_writable == old->host_writable &&
-           new->host_readable == old->host_readable;
-}
-
 static bool pl_buf_pool_grow(const struct pl_gpu *gpu, struct pl_buf_pool *pool)
 {
     const struct pl_buf *buf = pl_buf_create(gpu, &pool->current_params);
@@ -1131,7 +1115,7 @@ const struct pl_buf *pl_buf_pool_get(const struct pl_gpu *gpu,
 {
     pl_assert(!params->initial_data);
 
-    if (!pl_buf_params_compatible(params, &pool->current_params)) {
+    if (!pl_buf_params_superset(pool->current_params, *params)) {
         pl_buf_pool_uninit(gpu, pool);
         pool->current_params = *params;
     }
@@ -1228,7 +1212,6 @@ bool pl_tex_upload_texel(const struct pl_gpu *gpu, struct pl_dispatch *dp,
     pl_assert(params->buf);
     pl_assert(params->buf->params.type == PL_BUF_TEXEL_UNIFORM);
 
-    pl_dispatch_reset_frame(dp);
     struct pl_shader *sh = pl_dispatch_begin(dp);
     if (!sh_try_compute(sh, threads, 1, true, 0)) {
         PL_ERR(gpu, "Failed emulating texture transfer!");
@@ -1291,7 +1274,6 @@ bool pl_tex_download_texel(const struct pl_gpu *gpu, struct pl_dispatch *dp,
     pl_assert(params->buf);
     pl_assert(params->buf->params.type == PL_BUF_TEXEL_STORAGE);
 
-    pl_dispatch_reset_frame(dp);
     struct pl_shader *sh = pl_dispatch_begin(dp);
     if (!sh_try_compute(sh, threads, 1, true, 0)) {
         PL_ERR(gpu, "Failed emulating texture transfer!");
@@ -1368,12 +1350,26 @@ struct pl_pass_params pl_pass_params_copy(void *tactx,
     DUPSTRS(name, new.descriptors,    new.num_descriptors);
     DUPSTRS(name, new.vertex_attribs, new.num_vertex_attribs);
 
-    for (int i = 0; i < new.num_descriptors; i++) {
-        struct pl_desc *desc = &new.descriptors[i];
-        DUPSTRS(var.name, desc->buffer_vars, desc->num_buffer_vars);
-    }
-
 #undef DUPNAMES
 
     return new;
 }
+
+const struct pl_sync *pl_sync_create(const struct pl_gpu *gpu,
+                                     enum pl_handle_type handle_type)
+{
+    pl_assert(handle_type);
+    pl_assert(handle_type & gpu->handle_caps.sync);
+    pl_assert(PL_ISPOT(handle_type));
+    return gpu->impl->sync_create(gpu, handle_type);
+}
+
+void pl_sync_destroy(const struct pl_gpu *gpu,
+                     const struct pl_sync **sync)
+{
+    if (!*sync)
+        return;
+
+    gpu->impl->sync_destroy(gpu, *sync);
+    *sync = NULL;
+}
