@@ -32,9 +32,10 @@
 //
 // NOTE: When speaking of "valid usage" or "must", invalid usage is assumed to
 // result in undefined behavior. (Typically, an error message is printed to
-// stderr and libplacebo aborts). So ensuring valid API usage by the API user is
-// absolutely crucial. If you want to be freed from this reponsibility, use the
-// higher level abstractions provided by libplacebo alongside gpu.h.
+// stderr and libplacebo aborts - but this is not guaranteed). So ensuring
+// valid API usage by the API user is absolutely crucial. If you want to be
+// freed from this reponsibility, use the higher level abstractions provided by
+// libplacebo alongside gpu.h.
 
 // Structure which wraps metadata describing GLSL capabilities.
 struct pl_glsl_desc {
@@ -50,38 +51,73 @@ enum {
     PL_GPU_CAP_INPUT_VARIABLES  = 1 << 2, // supports shader input variables
 };
 
+// Some `pl_gpu` operations allow sharing GPU resources with external APIs -
+// examples include interop with other graphics APIs such as CUDA, and also
+// various hardware decoding APIs. This defines the mechanism underpinning the
+// communication of such an interoperation.
+typedef uint64_t pl_handle_caps;
+enum pl_handle_type {
+    PL_HANDLE_FD        = (1 << 0), // `int fd` for POSIX-style APIs
+    PL_HANDLE_WIN32     = (1 << 1), // `HANDLE` for win32 API
+    PL_HANDLE_WIN32_KMT = (1 << 2), // `HANDLE` for pre-Windows-8 win32 API
+};
+
+// Wrapper for the handle used to communicate a shared resource externally.
+// This handle is owned by the `pl_gpu` - if a user wishes to use it in a way
+// that takes over ownership (e.g. importing into some APIs), they must clone
+// the handle before doing so (e.g. using `dup` for fds). It is important to
+// read the external API documentation _very_ carefully as different handle
+// types may be managed in different ways. (eg: CUDA takes ownership of an fd,
+// but does not take ownership of a win32 handle).
+union pl_handle {
+    int fd;         // PL_HANDLE_FD
+    void *handle;   // PL_HANDLE_WIN32 / PL_HANDLE_WIN32_KMT
+};
+
+// Structure encapsulating memory that is shared between libplacebo and the
+// user. This memory can be imported into external APIs using the handle.
+//
+// If the object a `pl_shared_mem` belongs to is destroyed (e.g. via
+// `pl_buf_destroy`), the handle becomes undefined, as do the contents of the
+// memory it points to, as well as any external API objects imported from it.
+struct pl_shared_mem {
+    union pl_handle handle;
+    size_t size;   // the total size of the memory referenced by this handle
+    size_t offset; // the offset of the object within the referenced memory
+};
+
 // Structure defining the physical limits of this GPU instance. If a limit is
 // given as 0, that means that feature is unsupported.
 struct pl_gpu_limits {
-    int max_tex_1d_dim;    // maximum width for a 1D texture
-    int max_tex_2d_dim;    // maximum width/height for a 2D texture (required)
-    int max_tex_3d_dim;    // maximum width/height/depth for a 3D texture
-    size_t max_pushc_size; // maximum push_constants_size
-    size_t max_xfer_size;  // maximum size of a PL_BUF_TEX_TRANSFER
-    size_t max_ubo_size;   // maximum size of a PL_BUF_UNIFORM
-    size_t max_ssbo_size;  // maximum size of a PL_BUF_STORAGE
-    int max_buffer_texels; // maximum texels in a PL_BUF_TEXEL_*
-    int min_gather_offset; // minimum textureGatherOffset offset
-    int max_gather_offset; // maximum textureGatherOffset offset
+    uint32_t max_tex_1d_dim;    // maximum width for a 1D texture
+    uint32_t max_tex_2d_dim;    // maximum width/height for a 2D texture (required)
+    uint32_t max_tex_3d_dim;    // maximum width/height/depth for a 3D texture
+    size_t max_pushc_size;      // maximum `push_constants_size`
+    size_t max_xfer_size;       // maximum size of a PL_BUF_TEX_TRANSFER
+    size_t max_ubo_size;        // maximum size of a PL_BUF_UNIFORM
+    size_t max_ssbo_size;       // maximum size of a PL_BUF_STORAGE
+    uint64_t max_buffer_texels; // maximum texels in a PL_BUF_TEXEL_*
+    int16_t min_gather_offset;  // minimum `textureGatherOffset` offset
+    int16_t max_gather_offset;  // maximum `textureGatherOffset` offset
 
     // Compute shader limits. Always available (non-zero) if PL_GPU_CAP_COMPUTE set
-    size_t max_shmem_size; // maximum compute shader shared memory size
-    int max_group_threads; // maximum number of local threads per work group
-    int max_group_size[3]; // maximum work group size per dimension
-    int max_dispatch[3];   // maximum dispatch size per dimension
+    size_t max_shmem_size;      // maximum compute shader shared memory size
+    uint32_t max_group_threads; // maximum number of local threads per work group
+    uint32_t max_group_size[3]; // maximum work group size per dimension
+    uint32_t max_dispatch[3];   // maximum dispatch size per dimension
 
     // These don't represent hard limits but indicate performance hints for
     // optimal alignment. For best performance, the corresponding field
     // should be aligned to a multiple of these. They will always be a power
     // of two.
-    int align_tex_xfer_stride;    // optimal pl_tex_transfer_params.stride_w/h
-    size_t align_tex_xfer_offset; // optimal pl_tex_transfer_params.buf_offset
+    uint32_t align_tex_xfer_stride; // optimal `pl_tex_transfer_params.stride_w/h`
+    size_t align_tex_xfer_offset;   // optimal `pl_tex_transfer_params.buf_offset`
 };
 
 // Abstract device context which wraps an underlying graphics context and can
 // be used to dispatch rendering commands.
 struct pl_gpu {
-    struct pl_context *ctx;  // the pl_context this GPU was initialized from
+    struct pl_context *ctx;  // the `pl_context` this GPU was initialized from
     struct pl_gpu_fns *impl; // the underlying implementation (unique per GPU)
     void *priv;
 
@@ -91,6 +127,14 @@ struct pl_gpu {
     // Note: Every GPU must support at least one of PL_GPU_CAP_INPUT_VARIABLES
     // or uniform buffers (limits.max_ubo_size > 0).
 
+    // Fields relevant to external API interop. If the underlying device does
+    // not support interop with other APIs, these will all be {0}.
+    uint8_t uuid[16];               // underlying device UUID
+    struct {
+        pl_handle_caps shared_mem;  // supported handles for `pl_shared_mem`
+        pl_handle_caps sync;        // supported handles for `pl_sync`
+    } handle_caps;
+
     // Supported texture formats, in preference order. (If there are multiple
     // similar formats, the "better" ones come first)
     const struct pl_fmt **formats;
@@ -220,6 +264,11 @@ struct pl_tex_params {
     enum pl_tex_sample_mode sample_mode;
     enum pl_tex_address_mode address_mode;
 
+    // Setting this indicates that the memory backing this texture should be
+    // shared with external APIs, If so, this must be exactly *one* of
+    // `pl_gpu.handle_caps.shared_mem`.
+    enum pl_handle_type handle_type;
+
     // If non-NULL, the texture will be created with these contents. Using
     // this does *not* require setting host_writable. Otherwise, the initial
     // data is undefined.
@@ -246,6 +295,13 @@ static inline int pl_tex_params_dimension(const struct pl_tex_params params)
 struct pl_tex {
     struct pl_tex_params params;
     void *priv;
+
+    // If `params.handle_type` is set, this structure references the shared
+    // memory backing this buffer, via the requested handle type.
+    //
+    // While this texture is not in an "exported" state, the contents of the
+    // memory are undefined. (See: `pl_tex_export`)
+    struct pl_shared_mem shared_mem;
 };
 
 // Create a texture (with undefined contents). Returns NULL on failure. This is
@@ -256,6 +312,17 @@ const struct pl_tex *pl_tex_create(const struct pl_gpu *gpu,
 
 void pl_tex_destroy(const struct pl_gpu *gpu, const struct pl_tex **tex);
 
+// This works like `pl_tex_create`, but if the texture already exists and has
+// incompatible texture parameters, it will get destroyed first. A texture is
+// considered "compatible" if it has the same texture format and sample/address
+// mode and it supports a superset of the features the user requested.
+//
+// Even if the texture is not recreated, calling this function will still
+// invalidate the contents of the texture. (Note: Because of this,
+// `initial_data` may not be used with `pl_tex_recreate`. Doing so is an error)
+bool pl_tex_recreate(const struct pl_gpu *gpu, const struct pl_tex **tex,
+                     const struct pl_tex_params *params);
+
 // Invalidates the contents of a texture. After this, the contents are fully
 // undefined.
 void pl_tex_invalidate(const struct pl_gpu *gpu, const struct pl_tex *tex);
@@ -291,10 +358,11 @@ struct pl_tex_transfer_params {
 
     // Note: Superfluous parameters are ignored, i.e. for a 1D texture, the y
     // and z fields of `rc`, as well as the corresponding strides, are ignored.
-    // In all other cases, the stride must be >= the corresponding dimension
-    // of `rc`, and the `rc` must be normalized and fully contained within the
-    // image dimensions. If any of these parameters are left away (0), they
-    // are inferred from the texture's size.
+    // In all other cases, the stride must be >= the corresponding dimension of
+    // `rc`, and the `rc` must be normalized and fully contained within the
+    // image dimensions. Missing fields in the `rc` are inferred from the image
+    // size. If unset, the strides are inferred from `rc` (that is, it's
+    // assumed that the data is tightly packed in the buffer).
     struct pl_rect3d rc;   // region of the texture to transfer
     unsigned int stride_w; // the number of texels per horizontal row (x axis)
     unsigned int stride_h; // the number of texels per vertical column (y axis)
@@ -304,7 +372,9 @@ struct pl_tex_transfer_params {
     //
     // 1. Transferring to/from a buffer:
     const struct pl_buf *buf; // buffer to use (type must be PL_BUF_TEX_TRANSFER)
-    size_t buf_offset;        // offset of data within buffer, must be a multiple of 4
+    size_t buf_offset;        // offset of data within buffer, must be a
+                              // multiple of `tex->params.format->texel_size`
+                              // and should ideally be a multiple of 4
     // 2. Transferring to/from host memory directly:
     void *ptr;                // address of data
     // The contents of the memory region / buffer must exactly match the
@@ -338,15 +408,21 @@ bool pl_tex_download(const struct pl_gpu *gpu,
 // on a buffer.
 enum pl_buf_type {
     PL_BUF_INVALID = 0,
-    PL_BUF_TEX_TRANSFER, // texture transfer buffer (for pl_tex_upload/download)
-    PL_BUF_UNIFORM,      // UBO, for PL_DESC_BUF_UNIFORM
-    PL_BUF_STORAGE,      // SSBO, for PL_DESC_BUF_STORAGE
-    PL_BUF_TEXEL_UNIFORM,// texel buffer, for PL_DESC_BUF_TEXEL_UNIFORM
-    PL_BUF_TEXEL_STORAGE,// texel buffer, for PL_DESC_BUF_TEXEL_STORAGE
-    PL_BUF_PRIVATE,      // GPU-private usage (interpretation arbitrary)
+    PL_BUF_TEX_TRANSFER,  // texture transfer buffer (for pl_tex_upload/download)
+    PL_BUF_UNIFORM,       // UBO, for PL_DESC_BUF_UNIFORM
+    PL_BUF_STORAGE,       // SSBO, for PL_DESC_BUF_STORAGE
+    PL_BUF_TEXEL_UNIFORM, // texel buffer, for PL_DESC_BUF_TEXEL_UNIFORM
+    PL_BUF_TEXEL_STORAGE, // texel buffer, for PL_DESC_BUF_TEXEL_STORAGE
+    PL_BUF_PRIVATE,       // GPU-private usage (interpretation arbitrary)
     PL_BUF_TYPE_COUNT,
 };
 
+enum pl_buf_mem_type {
+    PL_BUF_MEM_AUTO = 0, // use whatever seems most appropriate
+    PL_BUF_MEM_HOST,     // try allocating from host memory (RAM)
+    PL_BUF_MEM_DEVICE,   // try allocating from device memory (VRAM)
+};
+
 // Structure describing a buffer.
 struct pl_buf_params {
     enum pl_buf_type type;
@@ -355,11 +431,22 @@ struct pl_buf_params {
     bool host_writable; // contents may be updated via pl_buf_write()
     bool host_readable; // contents may be read back via pl_buf_read()
 
+    // Provide a hint for the memory type you want to use when allocating
+    // this buffer's memory. Currently, this field is ignored for all buffer
+    // types except `PL_BUF_TEX_TRANSFER`, since uniform/storage buffers only
+    // make sense when allocated from device memory.
+    enum pl_buf_mem_type memory_type;
+
     // For texel buffers (PL_BUF_TEXEL_*), this gives the interpretation of the
     // buffer's contents. `format->caps` must include the corresponding
     // PL_FMT_CAP_TEXEL_* for the texel buffer type in use.
     const struct pl_fmt *format;
 
+    // Setting this indicates that the memory backing this buffer should be
+    // shared with external APIs, If so, this must be exactly *one* of
+    // `pl_gpu.handle_caps.shared_mem`.
+    enum pl_handle_type handle_type;
+
     // If non-NULL, the buffer will be created with these contents. Otherwise,
     // the initial data is undefined. Using this does *not* require setting
     // host_writable.
@@ -377,16 +464,43 @@ struct pl_buf_params {
 // vertex buffers, is designed to be completely fine.
 struct pl_buf {
     struct pl_buf_params params;
-    char *data; // for persistently mapped buffers, points to the first byte
+    uint8_t *data; // for persistently mapped buffers, points to the first byte
     void *priv;
+
+    // If `params.handle_type` is set, this structure references the shared
+    // memory backing this buffer, via the requested handle type.
+    //
+    // While this buffer is not in an "exported" state, the contents of the
+    // memory are undefined. (See: `pl_buf_export`)
+    struct pl_shared_mem shared_mem;
 };
 
 // Create a buffer. The type of buffer depends on the parameters. The buffer
 // parameters must adhere to the restrictions imposed by the pl_gpu_limits.
 // Returns NULL on failure.
+//
+// For buffers with shared memory, the buffer is considered to be in an
+// "exported" state by default, and may be used directly by the external API
+// after being created (until the first libplacebo operation on the buffer).
 const struct pl_buf *pl_buf_create(const struct pl_gpu *gpu,
                                    const struct pl_buf_params *params);
 
+// This behaves like `pl_buf_create`, but if the buffer already exists and has
+// incompatible parameters, it will get destroyed first. A buffer is considered
+// "compatible" if it has the same buffer type and texel format, a size greater
+// than or equal to the requested size, and it has a superset of the features
+// the user requested. After this operation, the contents of the buffer are
+// undefined.
+//
+// Note: Due to its unpredictability, it's not allowed to use this with
+// `params->initial_data` being set. Similarly, it's not allowed on a buffer
+// with `params->handle_type`. since this may invalidate the corresponding
+// external API's handle. Conversely, it *is* allowed on a buffer with
+// `params->host_mapped`, and the corresponding `buf->data` pointer *may*
+// change as a result of doing so.
+bool pl_buf_recreate(const struct pl_gpu *gpu, const struct pl_buf **buf,
+                     const struct pl_buf_params *params);
+
 void pl_buf_destroy(const struct pl_gpu *gpu, const struct pl_buf **buf);
 
 // Update the contents of a buffer, starting at a given offset (must be a
@@ -400,14 +514,49 @@ void pl_buf_write(const struct pl_gpu *gpu, const struct pl_buf *buf,
 bool pl_buf_read(const struct pl_gpu *gpu, const struct pl_buf *buf,
                  size_t buf_offset, void *dest, size_t size);
 
+// Initiates a buffer export operation, allowing a buffer to be accessed by an
+// external API. This is only valid for buffers with `params.handle_type`.
+// Calling this twice in a row is a harmless no-op. Returns whether successful.
+//
+// There is no corresponding "buffer import" operation, the next libplacebo
+// operation that touches the buffer (e.g. pl_tex_upload, but also pl_buf_write
+// and pl_buf_read) will implicitly import the buffer back to libplacebo. Users
+// must ensure that all pending operations made by the external API are fully
+// completed before using it in libplacebo again.
+//
+// Please note that this function returning does not mean the memory is
+// immediately available as such. In general, it will mark a buffer as "in use"
+// in the same way a read or write would, and it is the user's responsibility
+// to wait until `pl_buf_poll` returns false before accessing the memory from
+// the external API.
+//
+// In terms of the access performed by this operation, it is not considered a
+// "read" or "write" and therefore does not technically conflict with reads or
+// writes to the buffer performed by the host (via mapped memory - any use of
+// `pl_buf_read` or `pl_buf_write` would defeat the purpose of the export).
+// However, restrictions made by the external API may apply that prevent this.
+//
+// The recommended use pattern is something like this:
+//
+// while (loop) {
+//    const struct pl_buf *buf = get_free_buffer(); // or block on pl_buf_poll
+//    // write to the buffer using the external API
+//    pl_tex_upload(gpu, /* ... buf ... */); // implicitly imports
+//    pl_buf_export(gpu, buf);
+// }
+//
+// i.e. perform an external API operation, then use and immediately export the
+// buffer in libplacebo, and finally wait until `pl_buf_poll` is false before
+// re-using it. (Or get a new, fresh buffer in the meantime)
+bool pl_buf_export(const struct pl_gpu *gpu, const struct pl_buf *buf);
+
 // Returns whether or not a buffer is currently "in use". This can either be
-// because of a pending read operation or because of a pending write operation.
-// Coalescing multiple types of the same access (e.g. uploading the same buffer
-// to multiple textures) is fine, but trying to read a buffer while it is being
-// written to or trying to write to a buffer while it is being read from will
-// almost surely result in graphical corruption. The GPU makes no attempt to
-// enforce this, it is up to the user to check and adhere to whatever
-// restrictions are necessary.
+// because of a pending read operation, a pending write operation or a pending
+// buffer export operation. Any access to the buffer by the user is forbidden
+// while a buffer is "in use". This includes using `pl_buf_read` or
+// `pl_buf_write` or accessing mapped memory directly. The only exception to
+// this rule is multiple reads, for example reading from a buffer with
+// `pl_tex_upload` while simultaneously reading from it using mapped memory.
 //
 // The `timeout`, specified in nanoseconds, indicates how long to block for
 // before returning. If set to 0, this function will never block, and only
@@ -415,9 +564,16 @@ bool pl_buf_read(const struct pl_gpu *gpu, const struct pl_buf *buf,
 // timeout may be significantly longer than one nanosecond, and has no upper
 // bound. This function does not provide hard latency guarantees.
 //
-// Note: Destroying a buffer (pl_buf_destroy) is always valid, even if that
-// buffer is in use.
-bool pl_buf_poll(const struct pl_gpu *gpu, const struct pl_buf *buf, uint64_t timeout);
+// Note: Performing multiple libplacebo operations at the same time is always
+// valid, for example it's perfectly valid to submit a `pl_tex_upload`
+// immediately followed by a `pl_tex_download` to the same buffer. However,
+// when this is the case, it's undefined as to when exactly the upload is
+// happening versus when exactly the download is happening. So in this example,
+// any access to the buffer by the host would be forbidden, including reads,
+// until the user has verified that `pl_buf_poll` returned false. It's also
+// always valid to call `pl_buf_destroy`, even on in-use buffers.
+bool pl_buf_poll(const struct pl_gpu *gpu, const struct pl_buf *buf,
+                 uint64_t timeout);
 
 // Data type of a shader input variable (e.g. uniform, or UBO member)
 enum pl_var_type {
@@ -505,6 +661,8 @@ struct pl_var pl_var_from_fmt(const struct pl_fmt *fmt, const char *name);
 // distinction between the columns of a matrix and the rows of an array. For
 // example, a mat2[10] and a vec2[20] share the same pl_var_layout - the stride
 // would be sizeof(vec2) and the size would be sizeof(vec2) * 2 * 10.
+//
+// For non-array/matrix types, `stride` is equal to `size`.
 
 struct pl_var_layout {
     size_t offset; // the starting offset of the first byte
@@ -516,34 +674,23 @@ struct pl_var_layout {
 // tightly-packed, byte-aligned C data type, given a starting offset.
 struct pl_var_layout pl_var_host_layout(size_t offset, const struct pl_var *var);
 
-// Returns the layout requirements of a uniform buffer element given a current
-// buffer offset. If limits.max_ubo_size is 0, then this function returns {0}.
-//
-// Note: In terms of the GLSL, this is always *specified* as std140 layout, but
-// because of the way GLSL gets translated to other APIs (notably D3D11), the
-// actual buffer contents may vary considerably from std140. As such, the
-// calling code should not make any assumptions about the buffer layout and
-// instead query the layout requirements explicitly using this function.
+// Returns the GLSL std140 layout of an input variable given a current buffer
+// offset, as required for a buffer of type PL_BUF_UNIFORM.
 //
 // The normal way to use this function is when calculating the size and offset
 // requirements of a uniform buffer in an incremental fashion, to calculate the
 // new offset of the next variable in this buffer.
-struct pl_var_layout pl_buf_uniform_layout(const struct pl_gpu *gpu, size_t offset,
-                                           const struct pl_var *var);
+struct pl_var_layout pl_std140_layout(size_t offset, const struct pl_var *var);
 
-// Returns the layout requirements of a storage buffer element given a current
-// buffer offset. If limits.max_ssbo_size is 0, then this function returns {0}.
-//
-// Note: In terms of the GLSL, this is always *specified* as std430 layout, but
-// like with pl_buf_uniform_layout, the actual implementation may disagree.
-struct pl_var_layout pl_buf_storage_layout(const struct pl_gpu *gpu, size_t offset,
-                                           const struct pl_var *var);
+// Returns the GLSL std430 layout of an input variable given a current buffer
+// offset, as required for a buffer of type PL_BUF_STORAGE, and for push
+// constants.
+struct pl_var_layout pl_std430_layout(size_t offset, const struct pl_var *var);
 
-// Returns the layout requirements of a push constant element given a current
-// push constant offset. If `gpu->limits.max_pushc_size` is 0, then this
-// function returns {0}.
-struct pl_var_layout pl_push_constant_layout(const struct pl_gpu *gpu, size_t offset,
-                                             const struct pl_var *var);
+// Convenience definitions / friendly names for these
+#define pl_buf_uniform_layout pl_std140_layout
+#define pl_buf_storage_layout pl_std430_layout
+#define pl_push_constant_layout pl_std430_layout
 
 // Like memcpy, but copies bytes from `src` to `dst` in a manner governed by
 // the stride and size of `dst_layout` as well as `src_layout`. Also takes
@@ -593,11 +740,6 @@ enum pl_desc_access {
 // Returns the GLSL syntax for a given access mode (e.g. "readonly").
 const char *pl_desc_access_glsl_name(enum pl_desc_access mode);
 
-struct pl_buffer_var {
-    struct pl_var var;
-    struct pl_var_layout layout;
-};
-
 // Represents a shader descriptor (e.g. texture or buffer binding)
 struct pl_desc {
     const char *name;       // name as used in the shader
@@ -612,11 +754,6 @@ struct pl_desc {
     // the other descriptor types (uniform buffers and sampled textures are
     // always read-only).
     enum pl_desc_access access;
-
-    // For PL_DESC_BUF_UNIFORM/STORAGE, this specifies the layout of the
-    // variables contained by a buffer. Ignored for the other descriptor types
-    struct pl_buffer_var *buffer_vars;
-    int num_buffer_vars;
 };
 
 // Framebuffer blending mode (for raster passes)
@@ -695,8 +832,8 @@ struct pl_pass_params {
     // The target dummy texture this renderpass is intended to be used with.
     // This doesn't have to be a real texture - the caller can also pass a
     // blank pl_tex object, as long as target_dummy.params.format is set. The
-    // format must support PL_FMT_CAP_RENDERABLE, and the target dummy must
-    // have `renderable` enabled.
+    // format must support PL_FMT_CAP_RENDERABLE, and if the target dummy is
+    // an actual texture, it must have `renderable` enabled.
     //
     // If you pass a real texture here, the GPU backend may be able to optimize
     // the render pass better for the specific requirements of this texture.
@@ -767,8 +904,9 @@ struct pl_pass_run_params {
     // --- pass->params.type==PL_PASS_RASTER only
 
     // Target must be a 2D texture, target->params.renderable must be true, and
-    // target->params.format must match pass->params.target_fmt. If the viewport
-    // or scissors are left blank, they are inferred from target->params.
+    // target->params.format must match pass->params.target_dummy.params.format.
+    // If the viewport or scissors are left blank, they are inferred from
+    // target->params.
     //
     // WARNING: Rendering to a *target that is being read from by the same
     // shader is undefined behavior. In general, trying to bind the same
@@ -790,6 +928,59 @@ struct pl_pass_run_params {
 // Execute a render pass.
 void pl_pass_run(const struct pl_gpu *gpu, const struct pl_pass_run_params *params);
 
+// A generic synchronization object intended for use with an external API. This
+// is not required when solely using libplacebo API functions, as all required
+// synchronisation is done internally. This comes in the form of a pair of
+// semaphores - one to synchronize access in each direction.
+struct pl_sync {
+    enum pl_handle_type handle_type;
+    void *priv;
+
+    // This handle is signalled by the `pl_gpu`, and waited on by the user. It
+    // fires when it is safe for the user to access the shared resource.
+    union pl_handle wait_handle;
+
+    // This handle is signalled by the user, and waited on by the `pl_gpu`. It
+    // must fire when the user has finished accessing the shared resource.
+    union pl_handle signal_handle;
+};
+
+// Create a synchronization object. Returns NULL on failure.
+//
+// `handle_type` must be exactly *one* of `pl_gpu.handle_caps.sync`, and
+// indicates which type of handle to generate for sharing this sync object.
+const struct pl_sync *pl_sync_create(const struct pl_gpu *gpu,
+                                     enum pl_handle_type handle_type);
+
+// Destroy a `pl_sync`. Note that this invalidates the externally imported
+// semaphores. Users should therefore make sure that all operations that
+// wait on or signal any of the semaphore have been fully submitted and
+// processed by the external API before destroying the `pl_sync`.
+//
+// Despite this, it's safe to destroy a `pl_sync` if the only pending
+// operations that involve it are internal to libplacebo.
+void pl_sync_destroy(const struct pl_gpu *gpu,
+                     const struct pl_sync **sync);
+
+// Initiates a texture export operation, allowing a texture to be accessed by
+// an external API. This is only valid for textures with `params.handle_type`.
+// Returns whether successful. After this operation successfully returns, it is
+// guaranteed that `sync->wait_handle` will eventually be signalled. For APIs
+// where this is relevant, the image layout should be specified as "general",
+// e.g. `GL_LAYOUT_GENERAL_EXT` for OpenGL.
+//
+// There is no corresponding "import" operation - the next operation that uses
+// a texture will implicitly import the texture. Valid API usage requires that
+// the user *must* submit a semaphore signal operation on `sync->signal_handle`
+// before doing so. Not doing so is undefined behavior and may very well
+// deadlock the calling process and/or the graphics card!
+//
+// Note that despite this restriction, it is always valid to call
+// `pl_tex_destroy`, even if the texture is in an exported state, without
+// having to signal the corresponding sync object first.
+bool pl_tex_export(const struct pl_gpu *gpu, const struct pl_tex *tex,
+                   const struct pl_sync *sync);
+
 // This is semantically a no-op, but it provides a hint that you want to flush
 // any partially queued up commands and begin execution. There is normally no
 // need to call this, because queued commands will always be implicitly flushed
@@ -800,8 +991,28 @@ void pl_pass_run(const struct pl_gpu *gpu, const struct pl_pass_run_params *para
 // retroactively.
 //
 // The only time this might be beneficial to call explicitly is if you're doing
-// lots of offline rendering over a long period of time, and only fetching the
-// results (via pl_tex_download) at the very end.
+// lots of offline processing where you submit a bunch of work and then
+// use asynchronous texture downloads (via pl_tex_download) to retrieve the
+// results. In that case you should call this function after each work item
+// to ensure good parallelism between them.
+//
+// It's worth noting that this function may block if you're over-feeding the
+// GPU without waiting for existing results to finish.
 void pl_gpu_flush(const struct pl_gpu *gpu);
 
+// This is like `pl_gpu_flush` but also blocks until the GPU is fully idle
+// before returning. Using this in your rendering loop is seriously disadvised,
+// and almost never the right solution. The intended use case is for deinit
+// logic, where users may want to force the all pending GPU operations to
+// finish so they can clean up their state more easily.
+//
+// After this operation is called, it's guaranteed that all pending buffer
+// operations are complete - i.e. `pl_buf_poll` is guaranteed to return false.
+// Also, if you only care about buffer operations, you can accomplish this more
+// easily by using `pl_buf_poll` with the timeout set to `UINT64_MAX`. But
+// if you have many buffers it may be more convenient to call this function
+// instead. The difference is that this function will also affect e.g.
+// renders to a `pl_swapchain`.
+void pl_gpu_finish(const struct pl_gpu *gpu);
+
 #endif // LIBPLACEBO_GPU_H_
