@@ -19,6 +19,7 @@
 
 #include "common.h"
 #include "shaders.h"
+#include "dispatch.h"
 
 enum {
     // The scalers for each plane are set up to be just the index itself
@@ -311,7 +312,7 @@ static struct sampler_info sample_src_info(struct pl_renderer *rr,
 
     bool is_linear = src->tex->params.sample_mode == PL_TEX_SAMPLE_LINEAR;
     if (!rr->fbofmt || rr->disable_sampling || !info.config) {
-        info.type = is_linear ? SAMPLER_BICUBIC : SAMPLER_DIRECT;
+        info.type = SAMPLER_DIRECT;
     } else {
         info.type = SAMPLER_COMPLEX;
 
@@ -380,7 +381,7 @@ static void dispatch_sampler(struct pl_renderer *rr, struct pl_shader *sh,
     if (info.config->polar) {
         ok = pl_shader_sample_polar(sh, src, &fparams);
     } else {
-        struct pl_shader *tsh = pl_dispatch_begin(rr->dp);
+        struct pl_shader *tsh = pl_dispatch_begin_ex(rr->dp, true);
         ok = pl_shader_sample_ortho(tsh, PL_SEP_VERT, src, &fparams);
         if (!ok) {
             pl_dispatch_abort(rr->dp, &tsh);
@@ -544,10 +545,11 @@ static int deband_src(struct pl_renderer *rr, struct pl_shader *psh,
 
     struct pl_shader *sh = psh;
     struct pl_sample_src *src = psrc;
+    struct pl_sample_src fixed;
     if (!deband_scales) {
         // Only sample/deband the relevant cut-out, but round it to the nearest
         // integer to avoid doing fractional scaling
-        struct pl_sample_src fixed = *src;
+        fixed = *src;
         fixed.rect.x0 = floorf(fixed.rect.x0);
         fixed.rect.y0 = floorf(fixed.rect.y0);
         fixed.rect.x1 = ceilf(fixed.rect.x1);
@@ -564,7 +566,7 @@ static int deband_src(struct pl_renderer *rr, struct pl_shader *psh,
             // an exact integer crop without scaling), also skip the scalers
             deband_scales = true;
         } else {
-            sh = pl_dispatch_begin(rr->dp);
+            sh = pl_dispatch_begin_ex(rr->dp, true);
         }
     }
 
@@ -601,7 +603,7 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
                             const struct pl_image *image,
                             const struct pl_render_params *params)
 {
-    struct pl_shader *sh = pl_dispatch_begin(rr->dp);
+    struct pl_shader *sh = pl_dispatch_begin_ex(rr->dp, true);
     sh_require(sh, PL_SHADER_SIG_NONE, 0, 0);
 
     // Initialize the color to black
@@ -664,7 +666,7 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
     float scale = pl_color_repr_normalize(&repr);
 
     for (int i = 0; i < image->num_planes; i++) {
-        struct pl_shader *psh = pl_dispatch_begin(rr->dp);
+        struct pl_shader *psh = pl_dispatch_begin_ex(rr->dp, true);
         struct pl_plane *plane = &planes[i];
 
         // Compute the source shift/scale relative to the reference size
@@ -811,7 +813,7 @@ static bool pass_scale_main(struct pl_renderer *rr, struct pass_state *pass,
     draw_overlays(rr, src.tex, image->overlays, image->num_overlays,
                   img->color, use_sigmoid, NULL, params);
 
-    struct pl_shader *sh = pl_dispatch_begin(rr->dp);
+    struct pl_shader *sh = pl_dispatch_begin_ex(rr->dp, true);
     dispatch_sampler(rr, sh, &rr->samplers[SCALER_MAIN], params, &src);
     pass->cur_img = (struct img) {
         .sh     = sh,
@@ -893,6 +895,10 @@ fallback:
                             target->color, &rr->peak_detect_state, false);
     }
 
+    // Apply color blindness simulation if requested
+    if (params->cone_params)
+        pl_shader_cone_distort(sh, target->color, params->cone_params);
+
     bool is_comp = pl_shader_is_compute(sh);
     if (is_comp && !fbo->params.storable) {
         bool ok = finalize_img(rr, &pass->cur_img, rr->fbofmt, &rr->output_fbo);
