@@ -18,14 +18,13 @@
 #include <stdio.h>
 #include <math.h>
 
-#include "bstr/bstr.h"
 #include "common.h"
 #include "context.h"
 #include "shaders.h"
 
-struct pl_shader *pl_shader_alloc(struct pl_context *ctx,
-                                  const struct pl_gpu *gpu,
-                                  uint8_t ident, uint8_t index)
+struct pl_shader *pl_shader_alloc_ex(struct pl_context *ctx,
+                                     const struct pl_gpu *gpu,
+                                     uint8_t index, uint8_t ident)
 {
     pl_assert(ctx);
     struct pl_shader *sh = talloc_ptrtype(ctx, sh);
@@ -41,6 +40,12 @@ struct pl_shader *pl_shader_alloc(struct pl_context *ctx,
     return sh;
 }
 
+struct pl_shader *pl_shader_alloc(struct pl_context *ctx,
+                                  const struct pl_gpu *gpu, uint8_t index)
+{
+    return pl_shader_alloc_ex(ctx, gpu, index, 0);
+}
+
 void pl_shader_free(struct pl_shader **psh)
 {
     struct pl_shader *sh = *psh;
@@ -77,6 +82,11 @@ void pl_shader_reset(struct pl_shader *sh, uint8_t ident, uint8_t index)
     *sh = new;
 }
 
+bool pl_shader_is_failed(const struct pl_shader *sh)
+{
+    return sh->failed;
+}
+
 bool sh_try_compute(struct pl_shader *sh, int bw, int bh, bool flex, size_t mem)
 {
     pl_assert(bw && bh);
@@ -178,13 +188,13 @@ ident_t sh_attr_vec2(struct pl_shader *sh, const char *name,
                      const struct pl_rect2df *rc)
 {
     if (!sh->gpu) {
-        PL_ERR(sh, "Failed adding vertex attr '%s': No GPU available!", name);
+        SH_FAIL(sh, "Failed adding vertex attr '%s': No GPU available!", name);
         return NULL;
     }
 
     const struct pl_fmt *fmt = pl_find_vertex_fmt(sh->gpu, PL_FMT_FLOAT, 2);
     if (!fmt) {
-        PL_ERR(sh, "Failed adding vertex attr '%s': no vertex fmt!", name);
+        SH_FAIL(sh, "Failed adding vertex attr '%s': no vertex fmt!", name);
         return NULL;
     }
 
@@ -213,12 +223,12 @@ ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
                 ident_t *out_pos, ident_t *out_size, ident_t *out_pt)
 {
     if (!sh->gpu) {
-        PL_ERR(sh, "Failed binding texture '%s': No GPU available!", name);
+        SH_FAIL(sh, "Failed binding texture '%s': No GPU available!", name);
         return NULL;
     }
 
     if (pl_tex_params_dimension(tex->params) != 2 || !tex->params.sampleable) {
-        PL_ERR(sh, "Failed binding texture '%s': incompatible params!", name);
+        SH_FAIL(sh, "Failed binding texture '%s': incompatible params!", name);
         return NULL;
     }
 
@@ -260,6 +270,44 @@ ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
     return itex;
 }
 
+bool sh_buf_desc_append(void *tactx, const struct pl_gpu *gpu,
+                        struct pl_shader_desc *buf_desc,
+                        struct pl_var_layout *out_layout,
+                        const struct pl_var new_var)
+{
+    struct pl_buffer_var bv = { .var = new_var };
+    size_t cur_size = sh_buf_desc_size(buf_desc);
+
+    switch (buf_desc->desc.type) {
+    case PL_DESC_BUF_UNIFORM:
+        bv.layout = pl_std140_layout(cur_size, &new_var);
+        if (bv.layout.offset + bv.layout.size > gpu->limits.max_ubo_size)
+            return false;
+        break;
+    case PL_DESC_BUF_STORAGE:
+        bv.layout = pl_std430_layout(cur_size, &new_var);
+        if (bv.layout.offset + bv.layout.size > gpu->limits.max_ssbo_size)
+            return false;
+        break;
+    default: abort();
+    }
+
+    if (out_layout)
+        *out_layout = bv.layout;
+    TARRAY_APPEND(tactx, buf_desc->buffer_vars, buf_desc->num_buffer_vars, bv);
+    return true;
+}
+
+size_t sh_buf_desc_size(const struct pl_shader_desc *buf_desc)
+{
+    if (!buf_desc->num_buffer_vars)
+        return 0;
+
+    const struct pl_buffer_var *last;
+    last = &buf_desc->buffer_vars[buf_desc->num_buffer_vars - 1];
+    return last->layout.offset + last->layout.size;
+}
+
 void pl_shader_append(struct pl_shader *sh, enum pl_shader_buf buf,
                       const char *fmt, ...)
 {
@@ -290,6 +338,11 @@ ident_t sh_subpass(struct pl_shader *sh, const struct pl_shader *sub)
 {
     pl_assert(sh->mutable);
 
+    if (sh->ident == sub->ident) {
+        SH_FAIL(sh, "Failed merging shaders: conflicting identifiers!");
+        return NULL;
+    }
+
     // Check for shader compatibility
     int res_w = PL_DEF(sh->output_w, sub->output_w),
         res_h = PL_DEF(sh->output_h, sub->output_h);
@@ -297,8 +350,8 @@ ident_t sh_subpass(struct pl_shader *sh, const struct pl_shader *sub)
     if ((sub->output_w && res_w != sub->output_w) ||
         (sub->output_h && res_h != sub->output_h))
     {
-        PL_ERR(sh, "Failed merging shaders: incompatible sizes: %dx%d and %dx%d",
-               sh->output_w, sh->output_h, sub->output_w, sub->output_h);
+        SH_FAIL(sh, "Failed merging shaders: incompatible sizes: %dx%d and %dx%d",
+                sh->output_w, sh->output_h, sub->output_w, sub->output_h);
         return NULL;
     }
 
@@ -308,8 +361,8 @@ ident_t sh_subpass(struct pl_shader *sh, const struct pl_shader *sub)
         bool flex = sub->flexible_work_groups;
 
         if (!sh_try_compute(sh, subw, subh, flex, sub->res.compute_shmem)) {
-            PL_ERR(sh, "Failed merging shaders: incompatible block sizes or "
-                   "exceeded shared memory resource capabilities");
+            SH_FAIL(sh, "Failed merging shaders: incompatible block sizes or "
+                    "exceeded shared memory resource capabilities");
             return NULL;
         }
     }
@@ -360,6 +413,9 @@ static ident_t sh_split(struct pl_shader *sh)
 
 const struct pl_shader_res *pl_shader_finalize(struct pl_shader *sh)
 {
+    if (sh->failed)
+        return NULL;
+
     if (!sh->mutable) {
         PL_WARN(sh, "Attempted to finalize a shader twice?");
         return &sh->res;
@@ -380,17 +436,22 @@ const struct pl_shader_res *pl_shader_finalize(struct pl_shader *sh)
 
 bool sh_require(struct pl_shader *sh, enum pl_shader_sig insig, int w, int h)
 {
+    if (sh->failed) {
+        PL_ERR(sh, "Attempting to modify a failed shader!");
+        return false;
+    }
+
     if (!sh->mutable) {
-        PL_ERR(sh, "Attempted to modify an immutable shader!");
+        SH_FAIL(sh, "Attempted to modify an immutable shader!");
         return false;
     }
 
     if ((w && sh->output_w && sh->output_w != w) ||
         (h && sh->output_h && sh->output_h != h))
     {
-        PL_ERR(sh, "Illegal sequence of shader operations: Incompatible "
-               "output size requirements %dx%d and %dx%d",
-               sh->output_w, sh->output_h, w, h);
+        SH_FAIL(sh, "Illegal sequence of shader operations: Incompatible "
+                "output size requirements %dx%d and %dx%d",
+                sh->output_w, sh->output_h, w, h);
         return false;
     }
 
@@ -405,9 +466,9 @@ bool sh_require(struct pl_shader *sh, enum pl_shader_sig insig, int w, int h)
         pl_assert(!sh->res.input);
         sh->res.input = insig;
     } else if (sh->res.output != insig) {
-        PL_ERR(sh, "Illegal sequence of shader operations! Current output "
-               "signature is '%s', but called operation expects '%s'!",
-               names[sh->res.output], names[insig]);
+        SH_FAIL(sh, "Illegal sequence of shader operations! Current output "
+                "signature is '%s', but called operation expects '%s'!",
+                names[sh->res.output], names[insig]);
         return false;
     }
 
@@ -440,13 +501,13 @@ void *sh_require_obj(struct pl_shader *sh, struct pl_shader_obj **ptr,
 
     struct pl_shader_obj *obj = *ptr;
     if (obj && obj->gpu != sh->gpu) {
-        PL_ERR(sh, "Passed pl_shader_obj belongs to different GPU!");
+        SH_FAIL(sh, "Passed pl_shader_obj belongs to different GPU!");
         return NULL;
     }
 
     if (obj && obj->type != type) {
-        PL_ERR(sh, "Passed pl_shader_obj of wrong type! Shader objects must "
-               "always be used with the same type of shader.");
+        SH_FAIL(sh, "Passed pl_shader_obj of wrong type! Shader objects must "
+                "always be used with the same type of shader.");
         return NULL;
     }
 
@@ -479,9 +540,13 @@ ident_t sh_prng(struct pl_shader *sh, bool temporal, ident_t *p_state)
           "    return fract(state * 1.0/41.0);      \n"
           "}\n", permute, randfun, permute);
 
+    // Phi is the most irrational number, so it's a good candidate for
+    // generating seed values to the PRNG
+    static const double phi = 1.618033988749895;
+
     const char *seed = "0.0";
     if (temporal) {
-        float seedval = modff(M_PI * sh->index, &(float){0});
+        float seedval = modff(phi * sh->index, &(float){0});
         seed = sh_var(sh, (struct pl_shader_var) {
             .var  = pl_var_float("seed"),
             .data = &seedval,
@@ -490,9 +555,11 @@ ident_t sh_prng(struct pl_shader *sh, bool temporal, ident_t *p_state)
     }
 
     ident_t state = sh_fresh(sh, "prng");
-    GLSL("vec3 %s_m = vec3(gl_FragCoord.xy, %s) + vec3(1.0); \n"
-         "float %s = %s(%s(%s(%s_m.x) + %s_m.y) + %s_m.z);   \n",
-         state, seed,
+    GLSL("vec2 %s_init = fract(gl_FragCoord.xy * vec2(%f)); \n"
+         "vec3 %s_m = vec3(%s_init, %s) + vec3(1.0);        \n"
+         "float %s = %s(%s(%s(%s_m.x) + %s_m.y) + %s_m.z);  \n",
+         state, phi,
+         state, state, seed,
          state, permute, permute, permute, state, state, state);
 
     if (p_state)
@@ -589,12 +656,12 @@ next_dim: ; // `continue` out of the inner loop
                                     struct sh_lut_obj, sh_lut_uninit);
 
     if (!lut) {
-        PL_ERR(sh, "Failed initializing LUT object!");
+        SH_FAIL(sh, "Failed initializing LUT object!");
         goto error;
     }
 
     if (!gpu && method == SH_LUT_LINEAR) {
-        PL_ERR(sh, "Linear LUTs require the use of a GPU!");
+        SH_FAIL(sh, "Linear LUTs require the use of a GPU!");
         goto error;
     }
 
@@ -638,7 +705,7 @@ next_dim: ; // `continue` out of the inner loop
         case SH_LUT_TEXTURE:
         case SH_LUT_LINEAR: {
             if (!texdim) {
-                PL_ERR(sh, "Texture LUT exceeds texture dimensions!");
+                SH_FAIL(sh, "Texture LUT exceeds texture dimensions!");
                 goto error;
             }
 
@@ -653,7 +720,7 @@ next_dim: ; // `continue` out of the inner loop
             const struct pl_fmt *fmt;
             fmt = pl_find_fmt(gpu, PL_FMT_FLOAT, comps, 16, 32, caps);
             if (!fmt) {
-                PL_ERR(sh, "Found no compatible texture format for LUT!");
+                SH_FAIL(sh, "Found no compatible texture format for LUT!");
                 goto error;
             }
 
@@ -670,7 +737,7 @@ next_dim: ; // `continue` out of the inner loop
             });
 
             if (!lut->weights.tex) {
-                PL_ERR(sh, "Failed creating LUT texture!");
+                SH_FAIL(sh, "Failed creating LUT texture!");
                 goto error;
             }
             break;
@@ -715,11 +782,34 @@ next_dim: ; // `continue` out of the inner loop
     ident_t arr_name = NULL;
 
     static const char * const types[] = {"float", "vec2", "vec3", "vec4"};
+    static const char * const itypes[] = {"uint", "ivec2", "ivec3", "ivec4"};
     static const char * const swizzles[] = {"x", "xy", "xyz", "xyzw"};
 
     switch (method) {
-    case SH_LUT_TEXTURE:
+    case SH_LUT_TEXTURE: {
+        assert(texdim);
+        ident_t tex = sh_desc(sh, (struct pl_shader_desc) {
+            .desc = {
+                .name = "weights",
+                .type = PL_DESC_SAMPLED_TEX,
+            },
+            .object = lut->weights.tex,
+        });
+
+        GLSLH("#define %s(pos) (texelFetch(%s, %s(pos",
+              name, tex, itypes[texdim - 1]);
+
+        // Fill up extra components of the index
+        for (int i = dims; i < texdim; i++)
+            GLSLH(", 0");
+
+        GLSLH("), 0).%s)\n", swizzles[comps - 1]);
+        ret = name;
+        break;
+    }
+
     case SH_LUT_LINEAR: {
+        assert(texdim);
         ident_t tex = sh_desc(sh, (struct pl_shader_desc) {
             .desc = {
                 .name = "weights",
@@ -770,10 +860,10 @@ next_dim: ; // `continue` out of the inner loop
     }
 
     if (arr_name) {
-        GLSLH("#define %s(pos) (%s[int(%d * (pos).x)\\\n", name, arr_name, width);
+        GLSLH("#define %s(pos) (%s[int((pos).x)\\\n", name, arr_name);
         int shift = width;
         for (int i = 1; i < dims; i++) {
-            GLSLH("    + %d * int(%d * (pos)[%d])\\\n", shift, sizes[i], i);
+            GLSLH("    + %d * int((pos)[%d])\\\n", shift, i);
             shift *= sizes[i];
         }
         GLSLH("  ])\n");
