@@ -18,7 +18,6 @@
 #pragma once
 
 #include <stdio.h>
-#include "bstr/bstr.h"
 
 #include "common.h"
 #include "context.h"
@@ -43,6 +42,7 @@ struct pl_shader {
 
     // Internal state
     struct ta_ref *tmp;
+    bool failed;
     bool mutable;
     int output_w;
     int output_h;
@@ -55,6 +55,19 @@ struct pl_shader {
     int fresh;
 };
 
+#define SH_FAIL(sh, ...) do {    \
+        sh->failed = true;       \
+        PL_ERR(sh, __VA_ARGS__); \
+    } while (0)
+
+// Like `pl_shader_alloc`, but also has an extra `uint8_t ident` which can be
+// used to namespace shaders in order to allow safely merging together multiple
+// shaders using `sh_subpass`. This is not exposed publicly since there's no
+// reasonable public API for `sh_subpass`.
+struct pl_shader *pl_shader_alloc_ex(struct pl_context *ctx,
+                                     const struct pl_gpu *gpu,
+                                     uint8_t index, uint8_t ident);
+
 // Attempt enabling compute shaders for this pass, if possible
 bool sh_try_compute(struct pl_shader *sh, int bw, int bh, bool flex, size_t mem);
 
@@ -91,6 +104,19 @@ ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
                 const char *name, const struct pl_rect2df *rect,
                 ident_t *out_pos, ident_t *out_size, ident_t *out_pt);
 
+// Incrementally build up a buffer by adding new variable elements to the
+// buffer, resizing buf.buffer_vars if necessary. Returns whether or not the
+// variable could be successfully added (which may fail if you try exceeding
+// the size limits of the buffer type). If successful, the layout is stored
+// in *out_layout (may be NULL).
+bool sh_buf_desc_append(void *tactx, const struct pl_gpu *gpu,
+                        struct pl_shader_desc *buf_desc,
+                        struct pl_var_layout *out_layout,
+                        const struct pl_var new_var);
+
+size_t sh_buf_desc_size(const struct pl_shader_desc *buf_desc);
+
+
 // Underlying function for appending text to a shader
 void pl_shader_append(struct pl_shader *sh, enum pl_shader_buf buf,
                       const char *fmt, ...)
@@ -111,10 +137,10 @@ enum pl_shader_obj_type {
     PL_SHADER_OBJ_INVALID = 0,
     PL_SHADER_OBJ_PEAK_DETECT,
     PL_SHADER_OBJ_SAMPLER,
-    PL_SHADER_OBJ_SAMPLER_SEP,
     PL_SHADER_OBJ_DITHER,
     PL_SHADER_OBJ_3DLUT,
     PL_SHADER_OBJ_LUT,
+    PL_SHADER_OBJ_AV1_GRAIN,
 };
 
 struct pl_shader_obj {
@@ -155,9 +181,8 @@ enum sh_lut_method {
 // a given method (falling back if needed). The resulting identifier can be
 // sampled directly as %s(pos), where pos is a vector with the right number of
 // dimensions. `pos` must be an integer vector within the bounds of the array,
-// unless the method is `SH_LUT_LINEAR` or `SH_LUT_TEXTURE` in which case it's
-// a float vector that gets interpolated and clamped as needed. Returns NULL on
-// error.
+// unless the method is `SH_LUT_LINEAR`, in which case it's a float vector that
+// gets interpolated and clamped as needed. Returns NULL on error.
 //
 // This function also acts as `sh_require_obj`, and uses the `buf`, `tex`
 // and `text` fields of the resulting `obj`. (The other fields may be used by
