@@ -57,7 +57,8 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
 
     GLSL("color.rgb = %s * color.rgb + %s;\n", cmat, cmat_c);
 
-    if (orig_sys == PL_COLOR_SYSTEM_BT_2020_C) {
+    switch (orig_sys) {
+    case PL_COLOR_SYSTEM_BT_2020_C:
         // Conversion for C'rcY'cC'bc via the BT.2020 CL system:
         // C'bc = (B'-Y'c) / 1.9404  | C'bc <= 0
         //      = (B'-Y'c) / 1.5816  | C'bc >  0
@@ -90,6 +91,42 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
              "              1.0993 * pow(color.g, 0.45) - 0.0993, \n"
              "              %s(0.0181 <= color.g));               \n",
              sh_bvec(sh, 1));
+        break;
+
+    case PL_COLOR_SYSTEM_BT_2100_PQ:
+    case PL_COLOR_SYSTEM_BT_2100_HLG: {
+        // Conversion process from the spec:
+        //
+        // 1. L'M'S' = cmat * ICtCp
+        // 2. LMS = linearize(L'M'S')  (EOTF for PQ, inverse OETF for HLG)
+        // 3. RGB = lms2rgb * LMS
+        //
+        // After this we need to invert step 2 to arrive at non-linear RGB.
+        // (It's important we keep the transfer function conversion separate
+        // from the color system decoding, so we have to partially undo our
+        // work here even though we will end up linearizing later on anyway)
+        enum pl_color_transfer trc = orig_sys == PL_COLOR_SYSTEM_BT_2100_PQ
+                                        ? PL_COLOR_TRC_PQ
+                                        : PL_COLOR_TRC_HLG;
+
+        // Inverted from the matrix in the spec, transposed to column major
+        static const char *bt2100_lms2rgb = "mat3("
+            "  3.43661,  -0.79133, -0.0259499, "
+            " -2.50645,    1.9836, -0.0989137, "
+            "0.0698454, -0.192271,    1.12486) ";
+
+        pl_shader_linearize(sh, trc);
+        GLSL("color.rgb = %s * color.rgb; \n", bt2100_lms2rgb);
+        pl_shader_delinearize(sh, trc);
+        break;
+    }
+
+    case PL_COLOR_SYSTEM_XYZ:
+        break; // no special post-processing needed
+
+    default:
+        assert(pl_color_system_is_linear(orig_sys));
+        break;
     }
 
     if (repr->alpha == PL_ALPHA_INDEPENDENT) {
@@ -121,7 +158,7 @@ void pl_shader_encode_color(struct pl_shader *sh,
 
     if (!pl_color_system_is_linear(repr->sys)) {
         // FIXME: implement this case
-        PL_ERR(sh, "Non-linear color encoding currently unimplemented!");
+        SH_FAIL(sh, "Non-linear color encoding currently unimplemented!");
         return;
     }
 
@@ -480,13 +517,13 @@ static void sh_peak_uninit(const struct pl_gpu *gpu, void *ptr)
 static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
                             const struct pl_color_map_params *params)
 {
-    if (!state)
+    int frames = params->peak_detect_frames;
+    if (!state || !frames)
         return;
 
-    int frames = PL_DEF(params->peak_detect_frames, 63);
-    if (frames < 1 || frames > 1000) {
-        PL_ERR(sh, "Parameter peak_detect_frames must be >= 1 and <= 1000 "
-               "(was %d).", frames);
+    if (frames < 0 || frames > 1000) {
+        SH_FAIL(sh, "Parameter peak_detect_frames must be >= 0 and <= 1000 "
+                "(was %d).", frames);
         return;
     }
 
@@ -517,21 +554,22 @@ static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
     avg_total = pl_var_uint(sh_fresh(sh, "avg_total"));
 
     // Attempt packing the peak detection SSBO
-    struct pl_desc ssbo = {
-        .name   = "PeakDetect",
-        .type   = PL_DESC_BUF_STORAGE,
-        .access = PL_DESC_ACCESS_READWRITE,
+    struct pl_shader_desc ssbo = {
+        .desc = {
+            .name   = "PeakDetect",
+            .type   = PL_DESC_BUF_STORAGE,
+            .access = PL_DESC_ACCESS_READWRITE,
+        },
     };
 
-    struct pl_var_layout idx_l, num_l, ctr_l, max_l, avg_l, max_tl, avg_tl;
     bool ok = true;
-    ok &= pl_buf_desc_append(sh->tmp, gpu, &ssbo, &idx_l, idx);
-    ok &= pl_buf_desc_append(sh->tmp, gpu, &ssbo, &num_l, num);
-    ok &= pl_buf_desc_append(sh->tmp, gpu, &ssbo, &ctr_l, ctr);
-    ok &= pl_buf_desc_append(sh->tmp, gpu, &ssbo, &max_l, max);
-    ok &= pl_buf_desc_append(sh->tmp, gpu, &ssbo, &avg_l, avg);
-    ok &= pl_buf_desc_append(sh->tmp, gpu, &ssbo, &max_tl, max_total);
-    ok &= pl_buf_desc_append(sh->tmp, gpu, &ssbo, &avg_tl, avg_total);
+    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, idx);
+    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, num);
+    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, ctr);
+    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, max);
+    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, avg);
+    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, max_total);
+    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, avg_total);
 
     if (!ok) {
         PL_WARN(sh, "HDR peak detection exhausts device limits.. disabling");
@@ -540,7 +578,7 @@ static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
     }
 
     // Create the SSBO if necessary
-    size_t size = pl_buf_desc_size(&ssbo);
+    size_t size = sh_buf_desc_size(&ssbo);
     if (!obj->buf || obj->buf->params.size != size) {
         PL_TRACE(sh, "(Re)creating HDR peak detection SSBO");
 
@@ -548,22 +586,20 @@ static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
         pl_buf_destroy(gpu, &obj->buf);
         obj->buf = pl_buf_create(gpu, &(struct pl_buf_params) {
             .type = PL_BUF_STORAGE,
-            .size = pl_buf_desc_size(&ssbo),
+            .size = size,
             .initial_data = data,
         });
         talloc_free(data);
     }
 
     if (!obj->buf) {
-        PL_ERR(sh, "Failed creating peak detection SSBO!");
+        SH_FAIL(sh, "Failed creating peak detection SSBO!");
         return;
     }
 
     // Attach the SSBO and perform the peak detection logic
-    sh_desc(sh, (struct pl_shader_desc) {
-        .desc = ssbo,
-        .object = obj->buf,
-    });
+    ssbo.object = obj->buf;
+    sh_desc(sh, ssbo);
 
     // For performance, we want to do as few atomic operations on global
     // memory as possible, so use an atomic in shmem for the work group.
@@ -766,7 +802,7 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
 
     // Clip the final signal to the output range and apply the difference
     // linearly to the RGB channels. (this prevents discoloration)
-    GLSL("sig = min(sig, 1.0);        \n"
+    GLSL("sig = min(sig, 1.01);       \n"
         "color.rgb *= sig / sig_orig; \n");
 }
 
@@ -871,13 +907,13 @@ void pl_shader_color_map(struct pl_shader *sh,
     }
 
     // Tone map to rescale the signal average/peak if needed
-    if (src.sig_peak > dst.sig_peak)
+    if (src.sig_peak > dst.sig_peak + 1e-6)
         pl_shader_tone_map(sh, src, dst, dst_luma, peak_detect_state, params);
 
     // Warn for remaining out-of-gamut colors if enabled
     if (params->gamut_warning) {
-        GLSL("if (any(greaterThan(color.rgb, vec3(%f + 0.01))) ||\n"
-             "    any(lessThan(color.rgb, vec3(-0.01))))\n"
+        GLSL("if (any(greaterThan(color.rgb, vec3(%f + 0.005))) ||\n"
+             "    any(lessThan(color.rgb, vec3(-0.005))))\n"
              "    color.rgb = vec3(1.0) - color.rgb; // invert\n",
              dst.sig_peak);
     }
@@ -894,6 +930,32 @@ void pl_shader_color_map(struct pl_shader *sh,
     GLSL("}\n");
 }
 
+void pl_shader_cone_distort(struct pl_shader *sh, struct pl_color_space csp,
+                            const struct pl_cone_params *params)
+{
+    if (!params || !params->cones)
+        return;
+
+    if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
+        return;
+
+    GLSL("// pl_shader_cone_distort\n");
+    GLSL("{\n");
+
+    pl_color_space_infer(&csp);
+    pl_shader_linearize(sh, csp.transfer);
+
+    struct pl_matrix3x3 cone_mat;
+    cone_mat = pl_get_cone_matrix(params, pl_raw_primaries_get(csp.primaries));
+    GLSL("color.rgb = %s * color.rgb;\n", sh_var(sh, (struct pl_shader_var) {
+        .var = pl_var_mat3("cone_mat"),
+        .data = PL_TRANSPOSE_3X3(cone_mat.m),
+    }));
+
+    pl_shader_delinearize(sh, csp.transfer);
+    GLSL("}\n");
+}
+
 struct sh_dither_obj {
     enum pl_dither_method method;
     struct pl_shader_obj *lut;
@@ -956,7 +1018,7 @@ void pl_shader_dither(struct pl_shader *sh, int new_depth,
 
     params = PL_DEF(params, &pl_dither_default_params);
     if (params->lut_size < 0 || params->lut_size > 8) {
-        PL_ERR(sh, "Invalid `lut_size` specified: %d", params->lut_size);
+        SH_FAIL(sh, "Invalid `lut_size` specified: %d", params->lut_size);
         return;
     }
 
@@ -1054,7 +1116,7 @@ done: ;
 
     default: // LUT-based methods
         pl_assert(lut);
-        GLSL("bias = %s(pos);\n", lut);
+        GLSL("bias = %s(ivec2(pos * %d));\n", lut, lut_size);
         break;
     }
 
@@ -1067,6 +1129,7 @@ done: ;
 
 const struct pl_dither_params pl_dither_default_params = {
     .method     = PL_DITHER_BLUE_NOISE,
+    .lut_size   = 6,
     .temporal   = false, // commonly flickers on LCDs
 };
 
@@ -1156,7 +1219,7 @@ void pl_3dlut_apply(struct pl_shader *sh, struct pl_shader_obj **lut3d)
     obj = SH_OBJ(sh, lut3d, PL_SHADER_OBJ_3DLUT,
                  struct sh_3dlut_obj, sh_3dlut_uninit);
     if (!obj || !obj->lut || !obj->updated || !obj->ok) {
-        PL_ERR(sh, "pl_shader_3dlut called without prior pl_3dlut_update?");
+        SH_FAIL(sh, "pl_shader_3dlut called without prior pl_3dlut_update?");
         return;
     }
 
