@@ -71,7 +71,7 @@ void pl_shader_deband(struct pl_shader *sh, const struct pl_sample_src *src,
                       const struct pl_deband_params *params)
 {
     if (src->tex->params.sample_mode != PL_TEX_SAMPLE_LINEAR) {
-        PL_ERR(sh, "Debanding requires sample_mode = PL_TEX_SAMPLE_LINEAR!");
+        SH_FAIL(sh, "Debanding requires sample_mode = PL_TEX_SAMPLE_LINEAR!");
         return;
     }
 
@@ -169,8 +169,8 @@ static void bicubic_calcweights(struct pl_shader *sh, const char *t, const char
 bool pl_shader_sample_bicubic(struct pl_shader *sh, const struct pl_sample_src *src)
 {
     if (src->tex->params.sample_mode != PL_TEX_SAMPLE_LINEAR) {
-        PL_ERR(sh, "Trying to use fast bicubic sampling from a texture without "
-               "PL_TEX_SAMPLE_LINEAR");
+        SH_FAIL(sh, "Trying to use fast bicubic sampling from a texture without "
+                "PL_TEX_SAMPLE_LINEAR");
         return false;
     }
 
@@ -274,12 +274,14 @@ static void polar_sample(struct pl_shader *sh, const struct pl_filter *filter,
 struct sh_sampler_obj {
     const struct pl_filter *filter;
     struct pl_shader_obj *lut;
+    struct pl_shader_obj *pass2; // for pl_shader_sample_ortho
 };
 
 static void sh_sampler_uninit(const struct pl_gpu *gpu, void *ptr)
 {
     struct sh_sampler_obj *obj = ptr;
     pl_shader_obj_destroy(&obj->lut);
+    pl_shader_obj_destroy(&obj->pass2);
     pl_filter_free(&obj->filter);
     *obj = (struct sh_sampler_obj) {0};
 }
@@ -299,7 +301,7 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
 {
     pl_assert(params);
     if (!params->filter.polar) {
-        PL_ERR(sh, "Trying to use polar sampling with a non-polar filter?");
+        SH_FAIL(sh, "Trying to use polar sampling with a non-polar filter?");
         return false;
     }
 
@@ -351,7 +353,7 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
 
         if (!obj->filter) {
             // This should never happen, but just in case ..
-            PL_ERR(sh, "Failed initializing polar filter!");
+            SH_FAIL(sh, "Failed initializing polar filter!");
             return false;
         }
     }
@@ -359,7 +361,7 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     ident_t lut = sh_lut(sh, &obj->lut, SH_LUT_LINEAR, lut_entries, 0, 0, 1,
                          update, obj, fill_polar_lut);
     if (!lut) {
-        PL_ERR(sh, "Failed initializing polar LUT!");
+        SH_FAIL(sh, "Failed initializing polar LUT!");
         return false;
     }
 
@@ -489,18 +491,6 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     return true;
 }
 
-struct sh_sampler_sep_obj {
-    struct pl_shader_obj *samplers[2];
-};
-
-static void sh_sampler_sep_uninit(const struct pl_gpu *gpu, void *ptr)
-{
-    struct sh_sampler_sep_obj *obj = ptr;
-    for (int i = 0; i < PL_ARRAY_SIZE(obj->samplers); i++)
-        pl_shader_obj_destroy(&obj->samplers[i]);
-    *obj = (struct sh_sampler_sep_obj) {0};
-}
-
 static void fill_ortho_lut(void *priv, float *data, int w, int h, int d)
 {
     const struct sh_sampler_obj *obj = priv;
@@ -516,7 +506,7 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
 {
     pl_assert(params);
     if (params->filter.polar) {
-        PL_ERR(sh, "Trying to use separated sampling with a polar filter?");
+        SH_FAIL(sh, "Trying to use separated sampling with a polar filter?");
         return false;
     }
 
@@ -556,18 +546,18 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
     // 2. After fixing the source for `setup_src`, we lose information about
     //    the scaling ratio of the other component. (Although this is only a
     //    minor reason and could easily be changed with some boilerplate)
-    struct sh_sampler_sep_obj *sepobj;
-    sepobj = SH_OBJ(sh, params->lut, PL_SHADER_OBJ_SAMPLER_SEP,
-                    struct sh_sampler_sep_obj, sh_sampler_sep_uninit);
-    if (!sepobj)
-        return false;
-
     struct sh_sampler_obj *obj;
-    obj = SH_OBJ(sh, &sepobj->samplers[pass], PL_SHADER_OBJ_SAMPLER,
+    obj = SH_OBJ(sh, params->lut, PL_SHADER_OBJ_SAMPLER,
                  struct sh_sampler_obj, sh_sampler_uninit);
     if (!obj)
         return false;
 
+    if (pass != 0) {
+        obj = SH_OBJ(sh, &obj->pass2, PL_SHADER_OBJ_SAMPLER,
+                     struct sh_sampler_obj, sh_sampler_uninit);
+        assert(obj);
+    }
+
     float inv_scale = 1.0 / ratio[pass];
     inv_scale = PL_MAX(inv_scale, 1.0);
 
@@ -590,7 +580,7 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
 
         if (!obj->filter) {
             // This should never happen, but just in case ..
-            PL_ERR(sh, "Failed initializing separated filter!");
+            SH_FAIL(sh, "Failed initializing separated filter!");
             return false;
         }
     }
@@ -600,7 +590,7 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
     ident_t lut = sh_lut(sh, &obj->lut, SH_LUT_LINEAR, width, lut_entries, 0, 4,
                          update, obj, fill_ortho_lut);
     if (!lut) {
-        PL_ERR(sh, "Failed initializing separated LUT!");
+        SH_FAIL(sh, "Failed initializing separated LUT!");
         return false;
     }
 
