@@ -193,7 +193,6 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     // Repeat this a few times to test the caching
     for (int i = 0; i < 10; i++) {
         printf("iteration %d\n", i);
-        pl_dispatch_reset_frame(dp);
         struct pl_shader *sh = pl_dispatch_begin(dp);
 
         // For testing, force the use of CS if possible
@@ -234,7 +233,6 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 
 #if PL_HAVE_LCMS
     // Test the use of 3DLUTs if available
-    pl_dispatch_reset_frame(dp);
     struct pl_shader *sh = pl_dispatch_begin(dp);
     pl_shader_sample_direct(sh, &(struct pl_sample_src) { .tex = src });
 
@@ -414,8 +412,33 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         .color          = pl_color_space_srgb,
     };
 
-    pl_tex_clear(gpu, fbo, (float[4]){0});
-    REQUIRE(pl_render_image(rr, &image, &target, NULL));
+    for (int i = 0; i < 10; i ++) {
+        pl_tex_clear(gpu, fbo, (float[4]){0});
+        REQUIRE(pl_render_image(rr, &image, &target, NULL));
+    }
+
+    // Test some custom params
+    struct pl_render_params params = pl_render_default_params;
+    params.upscaler = &pl_filter_ewa_lanczos;
+
+    params.color_adjustment = &(struct pl_color_adjustment) {
+        .brightness = 0.1,
+        .contrast = 0.9,
+        .saturation = 1.5,
+        .gamma = 0.8,
+    };
+
+    params.color_map_params = &(struct pl_color_map_params) {
+        .intent = PL_INTENT_SATURATION,
+        .tone_mapping_algo = PL_TONE_MAPPING_CLIP,
+        .hdr_simulation = true,
+        .gamut_warning = true,
+    };
+
+    for (int i = 0; i < 5; i++) {
+        pl_tex_clear(gpu, fbo, (float[4]){0});
+        REQUIRE(pl_render_image(rr, &image, &target, &params));
+    }
 
     fbo_data = malloc(fbo->params.w * fbo->params.h * sizeof(float[4]));
     REQUIRE(pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
@@ -434,6 +457,8 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         printf("\n");
     }
 
+    // TODO: embed a reference texture and ensure it matches
+
 error:
     free(fbo_data);
     pl_renderer_destroy(&rr);
