@@ -20,6 +20,16 @@
 #include "../common.h"
 #include "../context.h"
 
+#ifdef __unix__
+#define VK_HAVE_UNIX 1
+#endif
+
+#ifdef _WIN32
+#define VK_HAVE_WIN32 1
+#include <windows.h>
+#include <vulkan/vulkan_win32.h>
+#endif
+
 // Vulkan allows the optional use of a custom allocator. We don't need one but
 // mark this parameter with a better name in case we ever decide to change this
 // in the future. (And to make the code more readable)
@@ -44,6 +54,7 @@ struct vk_ctx {
     VkPhysicalDevice physd;
     VkPhysicalDeviceLimits limits;
     VkPhysicalDeviceFeatures features;
+    VkExtent3D transfer_alignment; // for pool_transfer
     VkDevice dev;
 
     // Enabled extensions
@@ -74,6 +85,18 @@ struct vk_ctx {
     struct vk_signal **signals;
     int num_signals;
 
+    // Instance-level function pointers
+    VK_FUN(vkGetPhysicalDeviceProperties2KHR);
+    VK_FUN(vkGetPhysicalDeviceImageFormatProperties2KHR);
+    VK_FUN(vkGetPhysicalDeviceExternalBufferPropertiesKHR);
+    VK_FUN(vkGetPhysicalDeviceExternalSemaphorePropertiesKHR);
+
     // Device-level function pointers
     VK_FUN(vkCmdPushDescriptorSetKHR);
+    VK_FUN(vkGetMemoryFdKHR);
+    VK_FUN(vkGetSemaphoreFdKHR);
+#ifdef VK_HAVE_WIN32
+    VK_FUN(vkGetMemoryWin32HandleKHR);
+    VK_FUN(vkGetSemaphoreWin32HandleKHR);
+#endif
 };
