@@ -25,6 +25,7 @@ const struct pl_vk_inst_params pl_vk_inst_default_params = {0};
 struct vk_ext_fun {
     const char *name;
     size_t offset;
+    bool device_level;
 };
 
 struct vk_ext {
@@ -32,18 +33,32 @@ struct vk_ext {
     struct vk_ext_fun *funs;
 };
 
-#define VK_DEV_FUN(N)                       \
+#define VK_INST_FUN(N)                      \
     { .name = #N,                           \
       .offset = offsetof(struct vk_ctx, N), \
     }
 
-// Table of vulkan instance extensions
-// TODO: add support for loading function pointers
+#define VK_DEV_FUN(N)                       \
+    { .name = #N,                           \
+      .offset = offsetof(struct vk_ctx, N), \
+      .device_level = true,                 \
+    }
+
+// Table of optional vulkan instance extensions
 static const char *vk_instance_extensions[] = {
     VK_EXT_SWAPCHAIN_COLOR_SPACE_EXTENSION_NAME,
+    VK_KHR_EXTERNAL_MEMORY_CAPABILITIES_EXTENSION_NAME,
+    VK_KHR_EXTERNAL_SEMAPHORE_CAPABILITIES_EXTENSION_NAME,
 };
 
-// Table of vulkan device extensions and functions they load
+// Functions associated with mandatory instance extensions
+static const struct vk_ext_fun vk_inst_funs[] = {
+    VK_INST_FUN(vkGetPhysicalDeviceProperties2KHR),
+    VK_INST_FUN(vkGetPhysicalDeviceImageFormatProperties2KHR),
+};
+
+// Table of vulkan device extensions and functions they load, including
+// functions exported by dependent instance-level extensions
 static const struct vk_ext vk_device_extensions[] = {
     {
         .name = VK_KHR_PUSH_DESCRIPTOR_EXTENSION_NAME,
@@ -51,6 +66,46 @@ static const struct vk_ext vk_device_extensions[] = {
             VK_DEV_FUN(vkCmdPushDescriptorSetKHR),
             {0},
         },
+    }, {
+        .name = VK_KHR_EXTERNAL_MEMORY_EXTENSION_NAME,
+        .funs = (struct vk_ext_fun[]) {
+            VK_INST_FUN(vkGetPhysicalDeviceExternalBufferPropertiesKHR),
+            {0}
+        },
+    }, {
+        .name = VK_KHR_EXTERNAL_MEMORY_FD_EXTENSION_NAME,
+        .funs = (struct vk_ext_fun[]) {
+            VK_DEV_FUN(vkGetMemoryFdKHR),
+            {0},
+        },
+#ifdef VK_HAVE_WIN32
+    }, {
+        .name = VK_KHR_EXTERNAL_MEMORY_WIN32_EXTENSION_NAME,
+        .funs = (struct vk_ext_fun[]) {
+            VK_DEV_FUN(vkGetMemoryWin32HandleKHR),
+            {0},
+        },
+#endif
+    }, {
+        .name = VK_KHR_EXTERNAL_SEMAPHORE_EXTENSION_NAME,
+        .funs = (struct vk_ext_fun[]) {
+            VK_INST_FUN(vkGetPhysicalDeviceExternalSemaphorePropertiesKHR),
+            {0},
+        },
+    }, {
+        .name = VK_KHR_EXTERNAL_SEMAPHORE_FD_EXTENSION_NAME,
+        .funs = (struct vk_ext_fun[]) {
+            VK_DEV_FUN(vkGetSemaphoreFdKHR),
+            {0},
+        },
+#ifdef VK_HAVE_WIN32
+    }, {
+        .name = VK_KHR_EXTERNAL_SEMAPHORE_WIN32_EXTENSION_NAME,
+        .funs = (struct vk_ext_fun[]) {
+            VK_DEV_FUN(vkGetSemaphoreWin32HandleKHR),
+            {0},
+        },
+#endif
     }
 };
 
@@ -70,10 +125,11 @@ void pl_vk_inst_destroy(const struct pl_vk_inst **inst_ptr)
     TA_FREEP((void **) inst_ptr);
 }
 
-static VkBool32 vk_dbg_callback(VkDebugReportFlagsEXT flags,
-                                VkDebugReportObjectTypeEXT objType,
-                                uint64_t obj, size_t loc, int32_t msgCode,
-                                const char *layer, const char *msg, void *priv)
+static VkBool32 VKAPI_PTR vk_dbg_callback(VkDebugReportFlagsEXT flags,
+                                          VkDebugReportObjectTypeEXT objType,
+                                          uint64_t obj, size_t loc,
+                                          int32_t msgCode, const char *layer,
+                                          const char *msg, void *priv)
 {
     struct pl_context *ctx = priv;
     enum pl_log_level lev = PL_LOG_INFO;
@@ -162,13 +218,17 @@ const struct pl_vk_inst *pl_vk_inst_create(struct pl_context *ctx,
     info.ppEnabledExtensionNames = exts;
     info.enabledExtensionCount = num_exts;
 
+    pl_info(ctx, "Creating vulkan instance%s", num_exts ? " with extensions:" : "");
+    for (int i = 0; i < num_exts; i++)
+        pl_info(ctx, "    %s", exts[i]);
+
     VkResult res = vkCreateInstance(&info, VK_ALLOC, &inst);
     if (res != VK_SUCCESS) {
         pl_fatal(ctx, "Failed creating instance: %s", vk_res_str(res));
         goto error;
     }
 
-    VkDebugReportCallbackEXT debug = NULL;
+    VkDebugReportCallbackEXT debug = VK_NULL_HANDLE;
     if (params->debug) {
         // Set up a debug callback to catch validation messages
         VkDebugReportCallbackCreateInfoEXT dinfo = {
@@ -287,7 +347,7 @@ static bool find_physical_device(struct vk_ctx *vk,
             continue;
         }
 
-        if (params->device_name) {
+        if (params->device_name && params->device_name[0] != '\0') {
             if (strcmp(params->device_name, props.deviceName) == 0) {
                 vk->physd = devices[i];
                 best = 10; // high number...
@@ -426,6 +486,10 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     if (idx_comp < 0 && qfs[idx_gfx].queueFlags & VK_QUEUE_COMPUTE_BIT)
         idx_comp = idx_gfx;
 
+    // Cache the transfer queue alignment requirements
+    if (idx_tf >= 0)
+        vk->transfer_alignment = qfs[idx_tf].minImageTransferGranularity;
+
     // Now that we know which QFs we want, we can create the logical device
     VkDeviceQueueCreateInfo *qinfos = NULL;
     int num_qinfos = 0;
@@ -445,9 +509,13 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     if (params->surface)
         TARRAY_APPEND(vk, *exts, *num_exts, VK_KHR_SWAPCHAIN_EXTENSION_NAME);
 
-    // Add all extensions we can benefit from, and remember their functions
+    // Keep track of function pointers we need to load
     const struct vk_ext_fun **ext_funs = NULL;
     int num_ext_funs = 0;
+    for (int i = 0; i < PL_ARRAY_SIZE(vk_inst_funs); i++)
+        TARRAY_APPEND(tmp, ext_funs, num_ext_funs, &vk_inst_funs[i]);
+
+    // Add all optional device-level extensions extensions
     for (int i = 0; i < PL_ARRAY_SIZE(vk_device_extensions); i++) {
         const struct vk_ext *ext = &vk_device_extensions[i];
         for (int n = 0; n < num_exts_avail; n++) {
@@ -502,8 +570,12 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     // Load all of the functions from the extensions we enabled
     for (int i = 0; i < num_ext_funs; i++) {
         const struct vk_ext_fun *fun = ext_funs[i];
-        void *pfn = (void *) ((uintptr_t) vk + (ptrdiff_t) fun->offset);
-        *((PFN_vkVoidFunction *) pfn) = vkGetDeviceProcAddr(vk->dev, fun->name);
+        PFN_vkVoidFunction *pfn = (void *) ((uintptr_t) vk + (ptrdiff_t) fun->offset);
+        if (fun->device_level) {
+            *pfn = vkGetDeviceProcAddr(vk->dev, fun->name);
+        } else {
+            *pfn = vkGetInstanceProcAddr(vk->inst, fun->name);
+        };
     }
 
     // Create the command pools and memory allocator
@@ -545,7 +617,7 @@ const struct pl_vulkan *pl_vulkan_create(struct pl_context *ctx,
         pl_assert(!params->surface);
         pl_assert(!params->device);
         PL_DEBUG(vk, "No VkInstance provided, creating one...");
-        vk->internal_instance = pl_vk_inst_create(ctx, NULL);
+        vk->internal_instance = pl_vk_inst_create(ctx, params->instance_params);
         if (!vk->internal_instance)
             goto error;
         vk->inst = vk->internal_instance->instance;
