@@ -19,6 +19,10 @@
 #include "command.h"
 #include "utils.h"
 
+#ifdef VK_HAVE_UNIX
+#include <unistd.h>
+#endif
+
 // Controls the multiplication factor for new slab allocations. The new slab
 // will always be allocated such that the size of the slab is this factor times
 // the previous slab. Higher values make it grow faster.
@@ -62,18 +66,24 @@ struct vk_slab {
     struct vk_region *regions;
     int num_regions;
     // optional, depends on the memory type:
-    VkBuffer buffer;      // buffer spanning the entire slab
-    void *data;           // mapped memory corresponding to `mem`
+    VkBuffer buffer;        // buffer spanning the entire slab
+    void *data;             // mapped memory corresponding to `mem`
+    union pl_handle handle; // handle associated with this device memory
+    enum pl_handle_type handle_type;
 };
 
 // Represents a single memory heap. We keep track of a vk_heap for each
 // combination of buffer type and memory selection parameters. This shouldn't
 // actually be that many in practice, because some combinations simply never
 // occur, and others will generally be the same for the same objects.
+//
+// Note: `vk_heap` addresses are not immutable, so we musn't expose any dangling
+// references to a `vk_heap` from e.g. `vk_memslice.priv = vk_slab`.
 struct vk_heap {
     VkBufferUsageFlags usage;    // the buffer usage type (or 0)
     VkMemoryPropertyFlags flags; // the memory type flags (or 0)
     uint32_t typeBits;           // the memory type index requirements (or 0)
+    enum pl_handle_type handle_type; // handle type available for this heap
     struct vk_slab **slabs;      // array of slabs sorted by size
     int num_slabs;
 };
@@ -94,6 +104,17 @@ static void slab_free(struct vk_ctx *vk, struct vk_slab *slab)
 
     pl_assert(slab->used == 0);
     vkDestroyBuffer(vk->dev, slab->buffer, VK_ALLOC);
+
+#ifdef VK_HAVE_UNIX
+    if (slab->handle_type == PL_HANDLE_FD && slab->handle.fd > -1)
+        close(slab->handle.fd);
+#endif
+#ifdef VK_HAVE_WIN32
+    if (slab->handle_type == PL_HANDLE_WIN32 && slab->handle.handle != NULL)
+        CloseHandle(slab->handle.handle);
+    // PL_HANDLE_WIN32_KMT is just an identifier. It doesn't get closed.
+#endif
+
     // also implicitly unmaps the memory if needed
     vkFreeMemory(vk->dev, slab->mem, VK_ALLOC);
 
@@ -126,6 +147,29 @@ static bool find_best_memtype(struct vk_malloc *ma, uint32_t typeBits,
     return false;
 }
 
+static bool buf_export_check(struct vk_ctx *vk, VkBufferUsageFlags usage,
+                             enum pl_handle_type handle_type)
+{
+    if (!handle_type)
+        return true;
+
+    if (!vk->vkGetPhysicalDeviceExternalBufferPropertiesKHR)
+        return false;
+
+    VkPhysicalDeviceExternalBufferInfoKHR info = {
+        .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_BUFFER_INFO_KHR,
+        .usage = usage,
+        .handleType = vk_handle_type(handle_type),
+    };
+
+    VkExternalBufferPropertiesKHR props = {
+        .sType = VK_STRUCTURE_TYPE_EXTERNAL_BUFFER_PROPERTIES_KHR,
+    };
+
+    vk->vkGetPhysicalDeviceExternalBufferPropertiesKHR(vk->physd, &info, &props);
+    return vk_external_mem_check(&props.externalMemoryProperties, handle_type);
+}
+
 static struct vk_slab *slab_alloc(struct vk_malloc *ma, struct vk_heap *heap,
                                   size_t size)
 {
@@ -133,6 +177,7 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma, struct vk_heap *heap,
     struct vk_slab *slab = talloc_ptrtype(NULL, slab);
     *slab = (struct vk_slab) {
         .size = size,
+        .handle_type = heap->handle_type,
     };
 
     TARRAY_APPEND(slab, slab->regions, slab->num_regions, (struct vk_region) {
@@ -140,8 +185,24 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma, struct vk_heap *heap,
         .end   = slab->size,
     });
 
+    switch (slab->handle_type) {
+    case PL_HANDLE_FD:
+        slab->handle.fd = -1;
+        break;
+    case PL_HANDLE_WIN32:
+    case PL_HANDLE_WIN32_KMT:
+        slab->handle.handle = NULL;
+        break;
+    }
+
+    VkExportMemoryAllocateInfoKHR ext_info = {
+        .sType = VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO_KHR,
+        .handleTypes = vk_handle_type(slab->handle_type),
+    };
+
     VkMemoryAllocateInfo minfo = {
         .sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
+        .pNext = heap->handle_type ? &ext_info : NULL,
         .allocationSize = slab->size,
     };
 
@@ -155,8 +216,14 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma, struct vk_heap *heap,
         for (int i = 0; i < vk->num_pools; i++)
             qfs[i] = vk->pools[i]->qf;
 
+        VkExternalMemoryBufferCreateInfoKHR ext_buf_info = {
+            .sType = VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_BUFFER_CREATE_INFO_KHR,
+            .handleTypes = ext_info.handleTypes,
+        };
+
         VkBufferCreateInfo binfo = {
             .sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
+            .pNext = heap->handle_type ? &ext_buf_info : NULL,
             .size  = slab->size,
             .usage = heap->usage,
             .sharingMode = vk->num_pools > 1 ? VK_SHARING_MODE_CONCURRENT
@@ -165,6 +232,12 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma, struct vk_heap *heap,
             .pQueueFamilyIndices = qfs,
         };
 
+        if (!buf_export_check(vk, binfo.usage, slab->handle_type)) {
+            PL_ERR(vk, "Failed allocating shared memory buffer: possibly "
+                   "the handle type is unsupported?");
+            goto error;
+        }
+
         VK(vkCreateBuffer(vk->dev, &binfo, VK_ALLOC, &slab->buffer));
 
         VkMemoryRequirements reqs;
@@ -191,6 +264,33 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma, struct vk_heap *heap,
     if (slab->buffer)
         VK(vkBindBufferMemory(vk->dev, slab->buffer, slab->mem, 0));
 
+#ifdef VK_HAVE_UNIX
+    if (slab->handle_type == PL_HANDLE_FD) {
+        VkMemoryGetFdInfoKHR fd_info = {
+            .sType = VK_STRUCTURE_TYPE_MEMORY_GET_FD_INFO_KHR,
+            .memory = slab->mem,
+            .handleType = ext_info.handleTypes,
+        };
+
+        VK(vk->vkGetMemoryFdKHR(vk->dev, &fd_info, &slab->handle.fd));
+    }
+#endif
+
+#ifdef VK_HAVE_WIN32
+    if (slab->handle_type == PL_HANDLE_WIN32 ||
+        slab->handle_type == PL_HANDLE_WIN32_KMT)
+    {
+        VkMemoryGetWin32HandleInfoKHR handle_info = {
+            .sType = VK_STRUCTURE_TYPE_MEMORY_GET_WIN32_HANDLE_INFO_KHR,
+            .memory = slab->mem,
+            .handleType = ext_info.handleTypes,
+        };
+
+        VK(vk->vkGetMemoryWin32HandleKHR(vk->dev, &handle_info,
+                                         &slab->handle.handle));
+    }
+#endif
+
     return slab;
 
 error:
@@ -294,6 +394,24 @@ void vk_malloc_destroy(struct vk_malloc **ma_ptr)
     TA_FREEP(ma_ptr);
 }
 
+pl_handle_caps vk_malloc_handle_caps(struct vk_malloc *ma)
+{
+    struct vk_ctx *vk = ma->vk;
+    pl_handle_caps caps = 0;
+
+    for (int i = 0; vk_handle_list[i]; i++) {
+        // Try seeing if we could allocate a "basic" buffer using these
+        // capabilities, with no fancy buffer usage. More specific checks will
+        // happen down the line at VkBuffer creation time, but this should give
+        // us a rough idea of what the driver supports.
+        enum pl_handle_type type = vk_handle_list[i];
+        if (buf_export_check(vk, VK_BUFFER_USAGE_TRANSFER_DST_BIT, type))
+            caps |= type;
+    }
+
+    return caps;
+}
+
 void vk_free_memslice(struct vk_malloc *ma, struct vk_memslice slice)
 {
     struct vk_ctx *vk = ma->vk;
@@ -323,8 +441,10 @@ void vk_free_memslice(struct vk_malloc *ma, struct vk_memslice slice)
 // reqs: can be NULL
 static struct vk_heap *find_heap(struct vk_malloc *ma, VkBufferUsageFlags usage,
                                  VkMemoryPropertyFlags flags,
+                                 enum pl_handle_type handle_type,
                                  VkMemoryRequirements *reqs)
 {
+    pl_assert(PL_ISPOT(handle_type));
     int typeBits = reqs ? reqs->memoryTypeBits : 0;
 
     for (int i = 0; i < ma->num_heaps; i++) {
@@ -334,6 +454,8 @@ static struct vk_heap *find_heap(struct vk_malloc *ma, VkBufferUsageFlags usage,
             continue;
         if (ma->heaps[i].typeBits != typeBits)
             continue;
+        if (ma->heaps[i].handle_type != handle_type)
+            continue;
         return &ma->heaps[i];
     }
 
@@ -344,13 +466,14 @@ static struct vk_heap *find_heap(struct vk_malloc *ma, VkBufferUsageFlags usage,
         .usage    = usage,
         .flags    = flags,
         .typeBits = typeBits,
+        .handle_type = handle_type,
     };
     return heap;
 }
 
 static inline bool region_fits(struct vk_region r, size_t size, size_t align)
 {
-    return PL_ALIGN2(r.start, align) + size <= r.end;
+    return PL_ALIGN(r.start, align) + size <= r.end;
 }
 
 // Finds the best-fitting region in a heap. If the heap is too small or too
@@ -365,6 +488,7 @@ static bool heap_get_region(struct vk_malloc *ma, struct vk_heap *heap,
     // with the heap
     if (size > PLVK_HEAP_MAXIMUM_SLAB_SIZE) {
         slab = slab_alloc(ma, heap, size);
+        slab->dedicated = true;
         *out_slab = slab;
         *out_index = 0;
         return !!slab;
@@ -417,16 +541,22 @@ static bool slice_heap(struct vk_malloc *ma, struct vk_heap *heap, size_t size,
     struct vk_ctx *vk = ma->vk;
     struct vk_slab *slab;
     int index;
-    alignment = PL_ALIGN2(alignment, vk->limits.bufferImageGranularity);
+    alignment = pl_lcm(alignment, vk->limits.bufferImageGranularity);
     if (!heap_get_region(ma, heap, size, alignment, &slab, &index))
         return false;
 
     struct vk_region reg = slab->regions[index];
     TARRAY_REMOVE_AT(slab->regions, slab->num_regions, index);
+    VkDeviceSize offset = PL_ALIGN(reg.start, alignment);
     *out = (struct vk_memslice) {
         .vkmem = slab->mem,
-        .offset = PL_ALIGN2(reg.start, alignment),
+        .offset = offset,
         .size = size,
+        .shared_mem = {
+            .handle = slab->handle,
+            .offset = offset,
+            .size = slab->size,
+        },
         .priv = slab,
     };
 
@@ -442,17 +572,20 @@ static bool slice_heap(struct vk_malloc *ma, struct vk_heap *heap, size_t size,
 }
 
 bool vk_malloc_generic(struct vk_malloc *ma, VkMemoryRequirements reqs,
-                       VkMemoryPropertyFlags flags, struct vk_memslice *out)
+                       VkMemoryPropertyFlags flags,
+                       enum pl_handle_type handle_type,
+                       struct vk_memslice *out)
 {
-    struct vk_heap *heap = find_heap(ma, 0, flags, &reqs);
+    struct vk_heap *heap = find_heap(ma, 0, flags, handle_type, &reqs);
     return slice_heap(ma, heap, reqs.size, reqs.alignment, out);
 }
 
 bool vk_malloc_buffer(struct vk_malloc *ma, VkBufferUsageFlags bufFlags,
                       VkMemoryPropertyFlags memFlags, VkDeviceSize size,
-                      VkDeviceSize alignment, struct vk_bufslice *out)
+                      VkDeviceSize alignment, enum pl_handle_type handle_type,
+                      struct vk_bufslice *out)
 {
-    struct vk_heap *heap = find_heap(ma, bufFlags, memFlags, NULL);
+    struct vk_heap *heap = find_heap(ma, bufFlags, memFlags, handle_type, NULL);
     if (!slice_heap(ma, heap, size, alignment, &out->mem))
         return false;
 
