@@ -55,6 +55,10 @@ static bool vk_map_color_space(VkColorSpaceKHR space, struct pl_color_space *out
     // for any sort of typical SDR curve, which is better approximated by
     // `pl_color_space_monitor`.
     case VK_COLOR_SPACE_SRGB_NONLINEAR_KHR:
+        *out = pl_color_space_monitor;
+        return true;
+
+#ifdef VK_EXT_swapchain_colorspace
     case VK_COLOR_SPACE_BT709_NONLINEAR_EXT:
         *out = pl_color_space_monitor;
         return true;
@@ -122,6 +126,7 @@ static bool vk_map_color_space(VkColorSpaceKHR space, struct pl_color_space *out
     case VK_COLOR_SPACE_PASS_THROUGH_EXT:
         *out = pl_color_space_unknown;
         return true;
+#endif
 
     // Included to satisfy the switch coverage check
     case VK_COLOR_SPACE_RANGE_SIZE_KHR:
@@ -141,7 +146,7 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
 
     // Specific format requested by user
     if (out_format->format) {
-        if (vk_map_color_space(out_format->format, space)) {
+        if (vk_map_color_space(out_format->colorSpace, space)) {
             return true;
         } else {
             PL_ERR(gpu, "User-supplied surface format unsupported: 0x%x",
@@ -279,7 +284,7 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
     bool supported = false;
     for (int i = 0; i < num_modes; i++)
         supported |= (modes[i] == p->protoInfo.presentMode);
-    talloc_free(modes);
+    TA_FREEP(&modes);
 
     if (!supported) {
         PL_WARN(vk, "Requested swap mode unsupported by this device, falling "
@@ -334,6 +339,7 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
                         } alphaModes[] = {
         {VK_COMPOSITE_ALPHA_PRE_MULTIPLIED_BIT_KHR,  PL_ALPHA_PREMULTIPLIED},
         {VK_COMPOSITE_ALPHA_POST_MULTIPLIED_BIT_KHR, PL_ALPHA_INDEPENDENT},
+        {VK_COMPOSITE_ALPHA_INHERIT_BIT_KHR,         PL_ALPHA_UNKNOWN},
         {VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR,          PL_ALPHA_UNKNOWN},
     };
 
@@ -423,7 +429,7 @@ static void destroy_swapchain(struct vk_ctx *vk, struct priv *p)
 {
     assert(p->old_swapchain);
     vkDestroySwapchainKHR(vk->dev, p->old_swapchain, VK_ALLOC);
-    p->old_swapchain = NULL;
+    p->old_swapchain = VK_NULL_HANDLE;
 }
 
 static bool vk_sw_recreate(const struct pl_swapchain *sw)
@@ -516,7 +522,7 @@ error:
     PL_ERR(vk, "Failed (re)creating swapchain!");
     talloc_free(vkimages);
     vkDestroySwapchainKHR(vk->dev, p->swapchain, VK_ALLOC);
-    p->swapchain = NULL;
+    p->swapchain = VK_NULL_HANDLE;
     return false;
 }
 
@@ -534,7 +540,7 @@ static bool vk_sw_start_frame(const struct pl_swapchain *sw,
     for (int attempts = 0; attempts < 2; attempts++) {
         uint32_t imgidx = 0;
         VkResult res = vkAcquireNextImageKHR(vk->dev, p->swapchain, UINT64_MAX,
-                                             sem_in, NULL, &imgidx);
+                                             sem_in, VK_NULL_HANDLE, &imgidx);
 
         switch (res) {
         case VK_SUCCESS:
